import argparse
import sys
import zlib

from hsmk.cmds.hsmk import summarize_unsigned_spend
from hsmk.core.unsigned_spend import UnsignedSpend
from hsmk.util.qrint_encoding import a2b_qrint


def file_or_string(p) -> str:
    try:
        with open(p) as f:
            text = f.read().strip()
    except Exception:
        text = p
    return text


def fromhex_or_qrint(s: str) -> bytes:
    try:
        return a2b_qrint(s)
    except ValueError:
        pass
    return bytes.fromhex(s)


def hsmk_dump_us(args, parser):
    """
    Try to handle input in qrint or hex, with or without zlib compression
    """
    blob = fromhex_or_qrint(file_or_string(args.unsigned_spend))
    try:
        blob = zlib.decompress(blob)
    except zlib.error:
        pass
    unsigned_spend = UnsignedSpend.from_bytes(blob)
    summarize_unsigned_spend(unsigned_spend)


def create_parser():
    parser = argparse.ArgumentParser(
        description="Dump information about `UnsignedSpend`"
    )
    parser.add_argument(
        "unsigned_spend",
        metavar="hex-encoded-unsigned-spend-or-file",
        help="hex-encoded `UnsignedSpend`",
    )
    return parser


def main(argv=sys.argv[1:]):
    parser = create_parser()
    args = parser.parse_args(argv)
    return hsmk_dump_us(args, parser)


if __name__ == "__main__":  # pragma: no cover
    main()



"""
This file implements the `qrint` encoding of binary data to a string of decimal digits.

QR codes support different encoding modes, and generation tools typically choose them
automatically. There is a binary QR code encoding, but it doesn't work well with
hand-scanners since they act like keyboards, and most bytes can't be typed. So we need
to encode the data in a keyboard-friends subset of ASCII.

According to Wikipedia, the candidates are

- Kanji/Kana
- alphanumeric (0-9, A-Z, space plus 8 punctuation marks)
- binary (all 8-bit values)
- numeric (digits 0-9)

With western bias, I reject Kanji encoding as too difficult to read for
outsiders (and it probably hasn't had as much opportunity to be
debugged with hand-scanners).

Alphanumeric has an alphabet of 45 characters, many of them breaking,
which is an awkward number.

With binary, we are limited to the characters that can be easily typed,
which limits us to about 96 out of 256, which also places us fairly
awkwardly between standard methods that provide ease of encoding (like
base64) and efficiency (base96?).

Numeric is implemented by grouping the string of digits into 3 digit
groups and representing each as a ten-bit number from 0-1023. The
values 1000-1023 are unused, so we get some built-in inefficiency
giving us a theoretical maximum efficiency of `math.log(1000, 1024)` ~=
99.66%. Which actually isn't too bad, especially compared to base64,
which is 6 bits/8 bits = 75% efficient. But how do we convert an
integer to binary?

An obvious way is to use `int.from_bytes` and `int.to_bytes`. However,
converting long strings of digits to the bignum type required to hold
them gets slower non-linearly as the string length increases. So
instead we group the digits into groups of N digits, and figure out how
many bits we can cram into that value. For N=3, we get 9 bits, which
means for every ten bits, we would waste one (on top of the 0.34%
wasted already by QR numeric encoding). But we can do better by increasing N.

For example, N=63 is quite nice: the integer 1<<63=9223372036854775808 fits
in 64 bits and has 19 digits. Three blocks therefore has 57 digits, which
is encoded in a QR code using 190 bits, 189 of which represent actual data.
This yields an asymptotic efficiency of 99.47%, pretty darn close to the
theoretical maximum. One downside of using a large N is that the resulting
string length must be large even if there are few bytes, and we must include
information about how much padding is included (since beach block of 19 digits
add 63 bits to the stream).

Here is a table of good N values with their efficiency:

N  |  1<< N  | digits | bne3b | bui3b |  % (max 99.66%)
---+---------+--------+-------+-------+-----
3  |   8     |   1    |   9   |  10   |  90%
13 |   8192  |   4    |  39   |  40   |  97.5%
23 | 8388608 |   7    |  69   |  70   |  98.57%
33 | 85...92 |  10    |  99   | 100   |  99%
43 | 87...08 |  13    | 129   | 130   |  99.23%
53 | 90...92 |  16    | 159   | 160   |  99.375%
63 | 92...08 |  19    | 189   | 190   |  99.474%

bne3b = bits needed to express 3 blocks (3 * N)
bui3b = bits used (by QR code) in 3 blocks (10 * len(1<<N))

So let's standardize on N=3 or N=33, and add a prefix that indicates
which N value we're using and how many bytes of padding remain.

prefix 1: N=3, no padding (since this is effectively just the string in octal)
prefix 2: N=33, no padding
prefix 3: N=33, 1 byte padding
prefix 4: N=33, 2 bytes padding
prefix 5: N=33, 3 bytes padding
prefix 6: N=33, 4 bytes padding

An additional benefit of qrint encoding is that long integers don't
include breaking characters, making them easy to select with a mouse.
"""

from typing import Tuple

from chik_base.contrib.bech32m import convertbits


def b2a_qrint_payload(blob: bytes, grouping_size_bits: int) -> Tuple[int, str]:
    max_value = 1 << grouping_size_bits
    digit_block_size = len(str(max_value))
    block_count, extra_bits = divmod(len(blob) * 8, grouping_size_bits)

    if extra_bits > 0:
        block_count += 1

    bytes_in_block_count = block_count * grouping_size_bits // 8
    extra_bytes = bytes_in_block_count - len(blob)

    # now convert bytes into blocks of `grouping_size_bits` bits

    blocks = convertbits(blob, 8, grouping_size_bits, pad=True)
    format_template = "{:0%d}" % digit_block_size
    return extra_bytes, "".join(format_template.format(_) for _ in blocks)


def a2b_qrint_payload(s: str, grouping_size_bits: int) -> bytes:
    max_value = 1 << grouping_size_bits
    digit_block_size = len(str(max_value))

    len_s = len(s)
    blocks = [
        int(s[_ : _ + digit_block_size]) for _ in range(1, len_s, digit_block_size)
    ]
    return bytes(convertbits(blocks, grouping_size_bits, 8, pad=False))


def b2a_qrint(blob: bytes) -> str:
    MAX_SIZE_FOR_3_GROUP = 20

    padding_count, s33 = b2a_qrint_payload(blob, 33)

    if len(blob) < MAX_SIZE_FOR_3_GROUP:
        _, s3 = b2a_qrint_payload(blob, 3)
        if len(s3) < len(s33):
            return "1" + s3

    return "23456"[padding_count] + s33


PREFIX_TABLE = {
    "1": (3, 0),
    "2": (33, 0),
    "3": (33, 1),
    "4": (33, 2),
    "5": (33, 3),
    "6": (33, 4),
}


def a2b_qrint(s: str) -> bytes:
    c = s[0]
    if c not in PREFIX_TABLE:
        raise ValueError(f"illegal prefix {c}")
    grouping_size_bits, padding = PREFIX_TABLE[c]
    payload = a2b_qrint_payload(s, grouping_size_bits)
    if padding:
        payload = payload[:-padding]
    return payload



from dataclasses import dataclass
from typing import Iterable, List, Optional, Tuple
from weakref import WeakKeyDictionary

from chik_base.atoms import hexbytes
from chik_base.bls12_381 import BLSPublicKey, BLSSecretExponent
from chik_base.core import CoinSpend

from klvm_rs import Program  # type: ignore

from hsmk.core.signing_hints import SumHint, SumHints, PathHint, PathHints
from hsmk.core.unsigned_spend import SignatureInfo, UnsignedSpend
from hsmk.consensus.conditions import conditions_by_opcode
from hsmk.puzzles.conlang import AGG_SIG_ME, AGG_SIG_UNSAFE

MAX_COST = 1 << 34


@dataclass
class SignatureMetadata:
    partial_public_key: BLSPublicKey
    final_public_key: BLSPublicKey
    message: bytes


CONDITIONS_FOR_COIN_SPEND: WeakKeyDictionary = WeakKeyDictionary()


def conditions_for_coin_spend(coin_spend: CoinSpend) -> Program:
    if coin_spend not in CONDITIONS_FOR_COIN_SPEND:
        _cost, r = coin_spend.puzzle_reveal.run_with_cost(
            coin_spend.solution, max_cost=MAX_COST
        )
        CONDITIONS_FOR_COIN_SPEND[coin_spend] = r
    return CONDITIONS_FOR_COIN_SPEND[coin_spend]


def build_sum_hints_lookup(sum_hints: List[SumHint]) -> SumHints:
    return {_.final_public_key(): _ for _ in sum_hints}


def build_path_hints_lookup(path_hints: List[PathHint]) -> PathHints:
    return {_.public_key(): _ for _ in path_hints}


def sign(us: UnsignedSpend, secrets: List[BLSSecretExponent]) -> List[SignatureInfo]:
    sigs = []
    sum_hints = build_sum_hints_lookup(us.sum_hints)
    path_hints = build_path_hints_lookup(us.path_hints)
    for coin_spend in us.coin_spends:
        more_sigs = sign_for_coin_spend(
            coin_spend, secrets, sum_hints, path_hints, us.agg_sig_me_network_suffix
        )
        sigs.extend(more_sigs)
    return sigs


def sign_for_coin_spend(
    coin_spend: CoinSpend,
    secrets: List[BLSSecretExponent],
    sum_hints: SumHints,
    path_hints: PathHints,
    agg_sig_me_network_suffix: bytes,
) -> List[SignatureInfo]:
    conditions = conditions_for_coin_spend(coin_spend)
    agg_sig_me_message_suffix = coin_spend.coin.name() + agg_sig_me_network_suffix
    sigs = []
    for signature_metadata in partial_signature_metadata_for_hsm(
        conditions, sum_hints, path_hints, agg_sig_me_message_suffix
    ):
        partial_public_key = signature_metadata.partial_public_key
        final_public_key = signature_metadata.final_public_key
        message = signature_metadata.message
        path_hint = path_hints.get(partial_public_key) or PathHint(
            partial_public_key, []
        )
        secret_key = secret_key_for_public_key(
            secrets, path_hint.path, path_hint.root_public_key, partial_public_key
        )
        if secret_key is None:
            continue
        sig_info = SignatureInfo(
            secret_key.sign(message, final_public_key),
            partial_public_key,
            final_public_key,
            message,
        )

        sigs.append(sig_info)
    return sigs


def generate_synthetic_offset_signatures(us: UnsignedSpend) -> List[SignatureInfo]:
    sig_infos = []
    sum_hints = build_sum_hints_lookup(us.sum_hints)
    for coin_spend in us.coin_spends:
        for final_public_key, message in generate_verify_pairs(
            coin_spend, us.agg_sig_me_network_suffix
        ):
            sum_hint = sum_hints.get(final_public_key) or SumHint(
                [], BLSSecretExponent.zero()
            )

            secret_key = sum_hint.synthetic_offset
            partial_public_key = secret_key.public_key()
            signature = secret_key.sign(message, final_public_key)
            sig_info = SignatureInfo(
                signature, partial_public_key, final_public_key, message
            )
            sig_infos.append(sig_info)
    return sig_infos


def generate_verify_pairs(
    coin_spend: CoinSpend, agg_sig_me_network_suffix
) -> Iterable[Tuple[BLSPublicKey, bytes]]:
    agg_sig_me_message_suffix = coin_spend.coin.name() + agg_sig_me_network_suffix
    conditions = conditions_for_coin_spend(coin_spend)
    yield from verify_pairs_for_conditions(conditions, agg_sig_me_message_suffix)


def verify_pairs_for_conditions(
    conditions: Program, agg_sig_me_message_suffix: bytes
) -> Iterable[Tuple[BLSPublicKey, bytes]]:
    d = conditions_by_opcode(conditions)

    agg_sig_me_conditions = d.get(AGG_SIG_ME, [])
    for condition in agg_sig_me_conditions:
        yield (
            BLSPublicKey.from_bytes(condition.at("rf").atom),
            hexbytes(condition.at("rrf").atom + agg_sig_me_message_suffix),
        )

    agg_sig_unsafe_conditions = d.get(AGG_SIG_UNSAFE, [])
    for condition in agg_sig_unsafe_conditions:
        yield (
            BLSPublicKey.from_bytes(condition.at("rf").atom),
            hexbytes(condition.at("rrf").atom),
        )


def secret_key_for_public_key(
    secrets: List[BLSSecretExponent], path, root_public_key, public_key
) -> Optional[BLSSecretExponent]:
    for secret in secrets:
        if secret.public_key() == root_public_key:
            s = secret.child_for_path(path)
            if s.public_key() == public_key:
                return s
    return None


def partial_signature_metadata_for_hsm(
    conditions: Program,
    sum_hints: SumHints,
    path_hints: PathHints,
    agg_sig_me_message_suffix: bytes,
) -> Iterable[SignatureMetadata]:
    for final_public_key, message in verify_pairs_for_conditions(
        conditions,
        agg_sig_me_message_suffix,
    ):
        sum_hint = sum_hints.get(final_public_key) or SumHint(
            [final_public_key], BLSSecretExponent.zero()
        )

        for partial_public_key in sum_hint.public_keys:
            metadata = SignatureMetadata(partial_public_key, final_public_key, message)
            yield metadata



from dataclasses import dataclass, field
from typing import List, Optional, Union, Tuple

import random

import pytest

from chik_base.bls12_381 import BLSSecretExponent
from chik_base.core import Coin, CoinSpend
from chik_base.meta.typing import GenericAlias
from klvm_rs import Program  # type: ignore

from hsmk.klvm_serde import (
    from_program_for_type,
    to_program_for_type,
    tuple_frugal,
    EncodingError,
    Frugal,
)
from hsmk.core.signing_hints import SumHint, PathHint
from hsmk.core.unsigned_spend import (
    UnsignedSpend,
    from_storage,
    to_storage,
    TO_PROGRAM,
)
from .legacy.signing_hints import (
    SumHint as LegacySH,
    PathHint as LegacyPH,
)
from .legacy.unsigned_spend import UnsignedSpend as LegacyUS


def test_ser():
    tpb = to_program_for_type(bytes)
    fpb = from_program_for_type(bytes)
    tps = to_program_for_type(str)
    fps = from_program_for_type(str)
    for s in ["", "foo", "1" * 1000]:
        p = tps(s)
        assert p == Program.to(s)
        assert fps(p) == s

        b = s.encode()
        p = tpb(b)
        assert p == Program.to(b)
        assert fpb(p) == b

    tt = Tuple[str, bytes, int]
    tp = to_program_for_type(tt)
    fp = from_program_for_type(tt)
    for t in [
        ("foo", b"bar", 100),
        ("this is a test", bytes([5, 6, 7]), -94817),
    ]:
        assert tp(t) == Program.to(list(t))

    tt = List[Tuple[int, str]]
    tp = to_program_for_type(tt)
    fp = from_program_for_type(tt)
    for t in [
        [(100, "hundred"), (200, "two hundred"), (30, "thirty")],
    ]:
        rhs = list(list(_) for _ in t)
        prhs = Program.to(rhs)
        assert tp(t) == prhs
        assert fp(tp(t)) == t

    tt = GenericAlias(tuple_frugal, (int, str))
    tp = to_program_for_type(tt)
    fp = from_program_for_type(tt)
    for v in [
        (100, "hundred"),
        (200, "two hundred"),
        (30, "thirty"),
    ]:
        t = tuple_frugal(v)
        rhs = Program.to(t)
        assert tp(t) == rhs
        assert fp(tp(t)) == t

    @dataclass
    class Foo:
        a: int
        b: str

    tp = to_program_for_type(Foo)
    foo = Foo(100, "boss")
    rhs = Program.to([100, "boss"])
    assert tp(foo) == rhs
    fp = from_program_for_type(Foo)
    assert foo == fp(rhs)

    @dataclass
    class Nested:
        a: List[Foo]
        b: int

    tp = to_program_for_type(Nested)
    nested = Nested([foo, Foo(200, "worker")], 5000)
    rhs = Program.to([[[100, "boss"], [200, "worker"]], 5000])
    assert tp(nested) == rhs
    fp = from_program_for_type(Nested)
    assert nested == fp(tp(nested))

    @dataclass
    class Foo:
        a: int
        b: str = field(default="foo", metadata=dict(key="bob"))

    tp = to_program_for_type(Foo)
    foo = Foo(100, "boss")
    rhs = Program.to([100, ("bob", "boss")])
    assert tp(foo) == rhs
    fp = from_program_for_type(Foo)
    assert foo == fp(rhs)

    @dataclass
    class Foo:
        a: int = field(metadata=dict(key="a"))
        b: str = field(default="foo", metadata=dict(key="bob"))

    tp = to_program_for_type(Foo)
    fp = from_program_for_type(Foo)
    p = Program.to([("a", 1000), ("bob", "hello")])
    foo = fp(p)
    assert foo.a == 1000
    assert foo.b == "hello"
    p1 = tp(foo)
    assert p1 == p

    for foo in [Foo(20, "foo"), Foo(999, "bar"), Foo(-294, "baz")]:
        p = tp(foo)
        f1 = fp(p)
        assert f1 == foo

    @dataclass
    class Foo:
        a: Optional[int] = field(metadata=dict(key="a"))

    tp = to_program_for_type(Foo)
    fp = from_program_for_type(Foo)
    p = Program.to([("a", Program.to((1, 1000)))])
    foo = fp(p)
    assert foo.a == 1000
    p1 = tp(foo)
    assert p1 == p
    p = Program.to([("a", Program.to((Program.null(), Program.null())))])
    foo = fp(p)
    assert foo.a is None
    p1 = tp(foo)
    assert p1 == p

    @dataclass
    class Foo:
        a: Optional[List[int]] = field(metadata=dict(key="a"))

    tp = to_program_for_type(Foo)
    fp = from_program_for_type(Foo)
    assert fp(tp(Foo([]))) == Foo([])

    @dataclass
    class Bar:
        a: Union[int, str]

    with pytest.raises(ValueError):
        _ = to_program_for_type(Bar)

    with pytest.raises(ValueError):
        _ = from_program_for_type(Bar)


def test_serde_frugal():
    @dataclass
    class Foo(Frugal):
        a: int
        b: str

    tp = to_program_for_type(Foo)
    fp = from_program_for_type(Foo)
    p = Program.to((1000, "hello"))
    foo = fp(p)
    assert foo.a == 1000
    assert foo.b == "hello"
    p1 = tp(foo)
    assert p1 == p

    @dataclass
    class Bar(Frugal):
        a: int
        b: str
        c: List[int]

    tp = to_program_for_type(Bar)
    fp = from_program_for_type(Bar)
    p = Program.to((1000, ("hello", [5, 19, 220])))
    foo = fp(p)
    assert foo.a == 1000
    assert foo.b == "hello"
    assert foo.c == [5, 19, 220]
    p1 = tp(foo)
    assert p1 == p


def test_subclasses():
    # `chik-blockchain` has several int subclasses

    class Foo(int):
        pass

    tp = to_program_for_type(Foo)
    fp = from_program_for_type(Foo)
    for v in [-1000, -1, 0, 1, 100, 25678]:
        foo = Foo(v)
        p = Program.to(v)
        assert tp(foo) == p
        assert fp(p) == foo

    class Bar(bytes):
        pass

    tp = to_program_for_type(Bar)
    fp = from_program_for_type(Bar)
    for v in [b"", b"a", b"hello", b"a84kdhb8" * 500]:
        bar = Bar(v)
        p = Program.to(v)
        assert tp(bar) == p
        assert fp(p) == bar

    class Baz(str):
        pass

    tp = to_program_for_type(Baz)
    fp = from_program_for_type(Baz)
    for v in ["", "a", "hello", "a84kdhb8" * 500]:
        baz = Baz(v)
        p = Program.to(v)
        assert tp(baz) == p
        assert fp(p) == baz


def randbytes(r: random.Random, count: int) -> bytes:
    return bytes([r.randint(0, 255) for _ in range(count)])


def rnd_coin_spend(seed: int) -> CoinSpend:
    r = random.Random(seed)
    parent = randbytes(r, 32)
    puzzle = Program.to(f"puz: {randbytes(r, 5).hex()}")
    amount = r.randint(1, 1000) * int(1e3)
    solution = Program.to(f"sol: {randbytes(r, 10).hex()}")
    coin = Coin(parent, puzzle.tree_hash(), amount)
    return CoinSpend(coin, puzzle, solution)


def test_interop_sum_hint():
    pks = [BLSSecretExponent.from_int(_).public_key() for _ in range(5)]
    synthetic_offset = BLSSecretExponent.from_int(3**70 & ((1 << 256) - 1))
    sum_hint = SumHint(pks, synthetic_offset)
    tp = to_program_for_type(SumHint)
    fp = from_program_for_type(SumHint)
    p = tp(sum_hint)
    print(bytes(p).hex())
    lsh = LegacySH.from_program(p)
    print(lsh)
    assert lsh.public_keys == sum_hint.public_keys
    assert lsh.synthetic_offset == sum_hint.synthetic_offset
    assert lsh.final_public_key() == sum_hint.final_public_key()
    sh1 = fp(p)
    assert sum_hint == sh1


CoinSpendTuple = Tuple[bytes, Program, int, Program]


def test_interop_path_hint():
    public_key = BLSSecretExponent.from_int(1).public_key()
    ints = [1, 5, 91, 29484, 399]
    path_hint = PathHint(public_key, ints)
    tp = to_program_for_type(PathHint)
    fp = from_program_for_type(PathHint)
    p = tp(path_hint)
    print(bytes(p).hex())
    lph = LegacyPH.from_program(p)
    print(lph)
    assert lph.root_public_key == path_hint.root_public_key
    assert lph.path == path_hint.path
    assert lph.public_key() == path_hint.public_key()
    ph1 = fp(p)
    assert path_hint == ph1


def test_interop_coin_spend():
    cs_list = [rnd_coin_spend(_) for _ in range(10)]
    cst = from_storage(cs_list)
    cs_list_1 = to_storage(cst)
    assert cs_list == cs_list_1


def test_interop_unsigned_spend():
    cs_list = [rnd_coin_spend(_) for _ in range(10)]

    secret_keys = [BLSSecretExponent.from_int(_) for _ in range(5)]
    public_keys = [_.public_key() for _ in secret_keys]
    synthetic_offset = BLSSecretExponent.from_int(3**70 & ((1 << 256) - 1))
    sum_hint = SumHint(public_keys, synthetic_offset)
    lsh = LegacySH(public_keys, synthetic_offset)
    pubkey = BLSSecretExponent.from_int(9).public_key()
    ints = [2, 5, 17]
    path_hint = PathHint(pubkey, ints)
    lph = LegacyPH(pubkey, ints)

    suffix = b"a" * 32
    us = UnsignedSpend(cs_list[0:3], [sum_hint], [path_hint], suffix)

    lus = LegacyUS(cs_list[0:3], [lsh], [lph], suffix)
    print(bytes(lus.as_program()).hex())
    tp = to_program_for_type(UnsignedSpend)
    # TODO: remove this temporary hack
    # we add `__bytes__` to `UnsignedSpend` which changes how `to_program_for_type`
    # works
    tp = TO_PROGRAM
    fp = from_program_for_type(UnsignedSpend)
    p = tp(us)
    print(bytes(p).hex())
    lus = LegacyUS.from_program(p)
    assert lus.coin_spends == us.coin_spends
    for k in "public_keys synthetic_offset".split():
        for lh, rh in zip(lus.sum_hints, us.sum_hints):
            assert getattr(lh, k) == getattr(rh, k)
    for k in "root_public_key path".split():
        for lh, rh in zip(lus.path_hints, us.path_hints):
            assert getattr(lh, k) == getattr(rh, k)
    assert lus.agg_sig_me_network_suffix == us.agg_sig_me_network_suffix
    us1 = fp(p)
    assert us == us1


def test_tuple_frugal():
    Foo = GenericAlias(tuple_frugal, (int, str, bytes))

    tp = to_program_for_type(Foo)
    fp = from_program_for_type(Foo)
    p = Program.to((1000, ("hello", b"bob")))
    foo = fp(p)
    assert foo[0] == 1000
    assert foo[1] == "hello"
    assert foo[2] == b"bob"
    p1 = tp(foo)
    assert p1 == p


def test_dataclasses_transform():
    @dataclass
    class Foo:
        a: int = field(
            metadata=dict(alt_serde_type=(str, str, int), key="a"),
        )
        b: str = field(default="foo", metadata=dict(key="bob"))

    tp = to_program_for_type(Foo)
    fp = from_program_for_type(Foo)
    p = Program.to([("a", "1000"), ("bob", "hello")])
    foo = fp(p)
    assert foo.a == 1000
    assert foo.b == "hello"
    p1 = tp(foo)
    assert p1 == p


def test_failures():
    fp = from_program_for_type(bytes)
    with pytest.raises(EncodingError):
        fp(Program.to([1, 2]))

    tfi = GenericAlias(tuple_frugal, (int,))
    tp = to_program_for_type(tfi)
    with pytest.raises(EncodingError):
        tp((1, 2))

    fp = from_program_for_type(tfi)
    with pytest.raises(EncodingError):
        fp(Program.to((1, 2)))

    tp = to_program_for_type(Tuple[int])
    with pytest.raises(EncodingError):
        tp((1, 2))

    fp = from_program_for_type(Tuple[int])
    with pytest.raises(EncodingError):
        fp(Program.to([1, 2]))

    with pytest.raises(ValueError):
        from_program_for_type(object)

    with pytest.raises(ValueError):
        to_program_for_type(object)

    @dataclass
    class Foo:
        a: int = field(metadata=dict(key="a"))

    fp = from_program_for_type(Foo)
    with pytest.raises(EncodingError):
        fp(Program.to([]))



from typing import List

from chik_base.core import Coin
from chik_base.util.std_hash import std_hash

from klvm_rs import Program  # type: ignore

from hsmk.klvm.disasm import disassemble as bu_disassemble, KEYWORD_FROM_ATOM
from hsmk.consensus.conditions import conditions_by_opcode
from hsmk.process.sign import generate_verify_pairs
from hsmk.puzzles import conlang

KFA = {bytes([getattr(conlang, k)]): k for k in dir(conlang) if k[0] in "ACR"}

AGG_SIG_ME_ADDITIONAL_DATA = bytes.fromhex(
    "ccd5bb71183532bff220ba46c268991a3ff07eb358e8255a65c30a2dce0e5fbb"
)

MAX_COST = 1 << 34


# information needed to spend a cc
# if we ever support more genesis conditions, like a re-issuable coin,
# we may need also to save the `genesis_coin_mod` or its hash


def disassemble(sexp):
    """
    This version of `disassemble` also disassembles condition opcodes like
    `ASSERT_ANNOUNCEMENT_CONSUMED`.
    """
    kfa = dict(KEYWORD_FROM_ATOM)
    kfa.update((Program.to(k).atom, v) for k, v in KFA.items())
    return bu_disassemble(sexp, kfa)


def coin_as_program(coin: Coin) -> Program:
    """
    Convenience function for when putting `coin_info` into a solution.
    """
    return Program.to([coin.parent_coin_info, coin.puzzle_hash, coin.amount])


def dump_coin(coin: Coin) -> str:
    return disassemble(coin_as_program(coin))


def debug_spend_bundle(
    spend_bundle, agg_sig_additional_data=AGG_SIG_ME_ADDITIONAL_DATA
) -> None:
    """
    Print a lot of useful information about a `SpendBundle` that might help with
    debugging its klvm.
    """

    pks = []
    msgs = []

    created_coin_announcements: List[List[bytes]] = []
    asserted_coin_announcements = []
    created_puzzle_announcements: List[List[bytes]] = []
    asserted_puzzle_announcements = []

    print("=" * 80)
    for coin_spend in spend_bundle.coin_spends:
        coin = coin_spend.coin
        puzzle_reveal = Program.from_bytes(bytes(coin_spend.puzzle_reveal))
        solution = Program.from_bytes(bytes(coin_spend.solution))
        coin_name = coin.name()

        if puzzle_reveal.tree_hash() != coin_spend.coin.puzzle_hash:
            print("*** BAD PUZZLE REVEAL")
            print(
                f"{puzzle_reveal.tree_hash().hex()} vs"
                f" {coin_spend.coin.puzzle_hash.hex()}"
            )
            print("*" * 80)
            continue

        print(f"consuming coin {dump_coin(coin)}")
        print(f"  with id {coin_name}")
        print()
        print(
            f"\nbrun -y main.sym '{bu_disassemble(puzzle_reveal)}'"
            f" '{bu_disassemble(solution)}'"
        )
        cost, r = puzzle_reveal.run_with_cost(solution, max_cost=MAX_COST)
        conditions = conditions_by_opcode(r)
        error = None
        if error:
            print(f"*** error {error}")
        elif conditions is not None:
            for public_key, m in generate_verify_pairs(
                coin_spend, agg_sig_additional_data
            ):
                pks.append(public_key)
                msgs.append(m)
            print()
            print(disassemble(r))
            print(f"cost = {cost}")
            print()
            if conditions and len(conditions) > 0:
                print("grouped conditions:")
                for condition_programs in conditions.values():
                    print()
                    for c in condition_programs:
                        print(f"  {disassemble(Program.to(c))}")
                created_coin_announcements.extend(
                    [coin_name] + _.vars
                    for _ in conditions.get(conlang.CREATE_COIN_ANNOUNCEMENT, [])
                )
                asserted_coin_announcements.extend(
                    [
                        _.vars[0].hex()
                        for _ in conditions.get(conlang.ASSERT_COIN_ANNOUNCEMENT, [])
                    ]
                )
                created_puzzle_announcements.extend(
                    [puzzle_reveal.tree_hash()] + _.vars
                    for _ in conditions.get(conlang.CREATE_PUZZLE_ANNOUNCEMENT, [])
                )
                asserted_puzzle_announcements.extend(
                    [
                        _.vars[0].hex()
                        for _ in conditions.get(conlang.ASSERT_PUZZLE_ANNOUNCEMENT, [])
                    ]
                )
                print()
            else:
                print("(no output conditions generated)")
        print()
        print("-------")

    if 0:
        created = set(spend_bundle.additions())
        spent = set(spend_bundle.removals())

        zero_coin_set = set(coin.name() for coin in created if coin.amount == 0)

        ephemeral = created.intersection(spent)
        created.difference_update(ephemeral)
        spent.difference_update(ephemeral)
        print()
        print("spent coins")
        for coin in sorted(spent, key=lambda _: _.name()):
            print(f"  {dump_coin(coin)}")
            print(f"      => spent coin id {coin.name()}")
        print()
        print("created coins")
        for coin in sorted(created, key=lambda _: _.name()):
            print(f"  {dump_coin(coin)}")
            print(f"      => created coin id {coin.name()}")

        if ephemeral:
            print()
            print("ephemeral coins")
            for coin in sorted(ephemeral, key=lambda _: _.name()):
                print(f"  {dump_coin(coin)}")
                print(f"      => created coin id {coin.name()}")

        created_coin_announcement_pairs = [
            (_, std_hash(b"".join(_)).hex()) for _ in created_coin_announcements
        ]
        if created_coin_announcement_pairs:
            print("created coin announcements")
            for announcement, hashed in sorted(
                created_coin_announcement_pairs, key=lambda _: _[-1]
            ):
                as_hex = [f"0x{_.hex()}" for _ in announcement]
                print(f"  {as_hex} =>\n      {hashed}")

        eor_coin_announcements = sorted(
            set(_[-1] for _ in created_coin_announcement_pairs)
            ^ set(asserted_coin_announcements)
        )

        created_puzzle_announcement_pairs = [
            (_, std_hash(b"".join(_)).hex()) for _ in created_puzzle_announcements
        ]
        if created_puzzle_announcements:
            print("created puzzle announcements")
            for announcement, hashed in sorted(
                created_puzzle_announcement_pairs, key=lambda _: _[-1]
            ):
                as_hex = [f"0x{_.hex()}" for _ in announcement]
                print(f"  {as_hex} =>\n      {hashed}")

        eor_puzzle_announcements = sorted(
            set(_[-1] for _ in created_puzzle_announcement_pairs)
            ^ set(asserted_puzzle_announcements)
        )

        print()
        print()
        print(f"zero_coin_set = {sorted(zero_coin_set)}")
        print()
        if created_coin_announcement_pairs or asserted_coin_announcements:
            sa = sorted([_[-1] for _ in created_coin_announcement_pairs])
            print(f"created  coin announcements = {sa}")
            print()
            print(
                f"asserted coin announcements = {sorted(asserted_coin_announcements)}"
            )
            print()
            print(f"symdiff of coin announcements = {sorted(eor_coin_announcements)}")
            print()
        if created_puzzle_announcement_pairs or asserted_puzzle_announcements:
            sa = sorted([_[-1] for _ in created_puzzle_announcement_pairs])
            print(f"created  puzzle announcements = {sa}")
            print()
            print(
                f"asserted puzzle announcements ="
                f" {sorted(asserted_puzzle_announcements)}"
            )
            print()
            print(
                f"symdiff of puzzle announcements ="
                f" {sorted(eor_puzzle_announcements)}"
            )
            print()
    print()
    print("=" * 80)
    print()
    signature = spend_bundle.aggregated_signature
    validates = signature.verify(list(zip(pks, msgs)))
    print(f"aggregated signature check pass: {validates}")
    print(f"pks: {pks}")
    print(f"msgs: {[msg.hex() for msg in msgs]}")
    print(f"  msg_data: {[msg.hex()[:-128] for msg in msgs]}")
    print(f"  coin_ids: {[msg.hex()[-128:-64] for msg in msgs]}")
    print(f"  add_data: {[msg.hex()[-64:] for msg in msgs]}")
    print(f"signature: {signature}")
    return validates



from klvm_rs import Program

from hsmk.klvm.disasm import disassemble


def check_disassemble(h, s):
    p = Program.fromhex(h)
    output = disassemble(p)
    print(output)
    assert output == s


def test_disassemble():
    # one-byte encodings
    check_disassemble("80", "0")
    check_disassemble("01", "1")
    check_disassemble("3f", "63")

    check_disassemble("8140", "64")
    check_disassemble("817f", "127")
    check_disassemble("8180", "-128")
    check_disassemble("8181", "-127")
    check_disassemble("81fe", "-2")
    check_disassemble("81ff", "-1")
    check_disassemble("8100", "0x00")
    check_disassemble("8101", "1")
    check_disassemble("8102", "2")

    check_disassemble("820000", "0x0000")
    check_disassemble("820001", "0x0001")
    check_disassemble("820002", "0x0002")
    check_disassemble("82007f", "0x007f")
    check_disassemble("820080", "128")
    check_disassemble("820081", "129")
    check_disassemble("827fff", "32767")
    check_disassemble("828000", "-32768")
    check_disassemble("828001", "-32767")
    check_disassemble("82ff7f", "-129")
    check_disassemble("82ff80", "0xff80")
    check_disassemble("82ff81", "0xff81")
    check_disassemble("82fffe", "0xfffe")
    check_disassemble("82ffff", "0xffff")

    check_disassemble("83000000", "0x000000")
    check_disassemble("83000001", "0x000001")
    check_disassemble("83000002", "0x000002")
    check_disassemble("83007fff", "0x007fff")
    check_disassemble("83008000", "0x008000")
    check_disassemble("83008001", "0x008001")
    check_disassemble("837fffff", "0x7fffff")
    check_disassemble("83008000", "0x008000")
    check_disassemble("83008001", "0x008001")
    check_disassemble("8300ff7f", "0x00ff7f")
    check_disassemble("8300ff80", "0x00ff80")
    check_disassemble("8300ff81", "0x00ff81")
    check_disassemble("8300fffe", "0x00fffe")
    check_disassemble("8300ffff", "0x00ffff")

    check_disassemble("83666f6f", '"foo"')

    check_disassemble("ff8080", "(0)")
    check_disassemble("ff83666f6fff8362617280", '("foo" "bar")')
    check_disassemble(
        "ff83666f6fff83626172ff8362617a836a6f62", '("foo" "bar" "baz" . "job")'
    )
    check_disassemble(
        "ff83666f6fffff83626172ff8362617a80836a6f62", '("foo" ("bar" "baz") . "job")'
    )
    check_disassemble("ff64ff8200c880", "(100 200)")
    check_disassemble("ff01ff02ff05ff0d80", "(q 2 5 13)")
    check_disassemble("ff02ff05ff0d80", "(a 5 13)")

    sq, dq = "'", '"'
    # sq = single quote; dq = double quote
    ev = f"{sq}{dq}foo{dq}{sq}"
    check_disassemble("8522666f6f22", ev)

    # we now do the seven character string "'foo'"
    check_disassemble("872227666f6f2722", "0x2227666f6f2722")



# gratuitous coverage tests

import pathlib
import tempfile
import zlib

from klvm_rs import Program

from hsmk.process.sign import verify_pairs_for_conditions
from hsmk.puzzles.conlang import AGG_SIG_ME, AGG_SIG_UNSAFE
from hsmk.puzzles.p2_delegated_puzzle_or_hidden_puzzle import (
    calculate_synthetic_secret_key,
    solution_for_hidden_puzzle,
)
from hsmk.puzzles.p2_conditions import solution_for_conditions
from hsmk.util.byte_chunks import (
    blob_for_chunks,
    blob_for_zlib_chunks,
    create_chunks_for_blob,
    ChunkAssembler,
)

import pytest

from .generate import bytes32_generate, se_generate, pk_generate


def test_cssk():
    se = se_generate(1)
    b32 = bytes32_generate(1)
    xse = 2483436177918370406060013642567697118387289473841058299688804535472357425177
    assert int(calculate_synthetic_secret_key(se, b32)) == xse


def test_sfhp():
    pk = pk_generate(1)
    hp = Program.to("foobar")
    sthp = Program.to("fake solve")
    s = solution_for_hidden_puzzle(pk, hp, sthp)
    ethh = "eca8c5557fe1c4929f8d3dd7a5551d023c9a6c4215e69ebe3eb55f84b5a25c95"
    assert s.tree_hash().hex() == ethh


def test_sfc():
    s = solution_for_conditions(Program.to("a bunch of conditions"))
    ethh = "adabdc6a70925a06b1bd3e148d910793abf879de327adc1ff9ea405d72f29a48"
    assert s.tree_hash().hex() == ethh


def test_byte_chunks():
    blob = b"foo" * 10240
    with pytest.raises(ValueError):
        create_chunks_for_blob(blob, 5)

    chunks_1 = create_chunks_for_blob(blob, 2000)
    chunks_2 = create_chunks_for_blob(blob, 4000)
    with pytest.raises(ValueError):
        bogus_list = [chunks_1[0], chunks_2[0]]
        blob_for_chunks(bogus_list)

    with pytest.raises(ValueError):
        bogus_c1 = bytearray(chunks_1[0])
        bogus_c1[0] += 1
        bogus_c1 = bytes(bogus_c1)
        bogus_list = [chunks_1[0], bogus_c1]
        blob_for_chunks(bogus_list)

    ca = ChunkAssembler()
    assert ca.status() == (0, 0)

    c_blob = zlib.compress(blob)
    out = blob_for_zlib_chunks(create_chunks_for_blob(c_blob, 1000))
    assert out == blob

    with pytest.raises(ValueError):
        blob_for_chunks([chunks_1[0]])


def test_aggsig_me_conditions():
    pk_1 = pk_generate(1)
    pk_2 = pk_generate(2)
    b32_1 = bytes32_generate(1)
    b32_2 = bytes32_generate(2)
    b32_3 = bytes32_generate(3)
    conditions = Program.to(
        [[AGG_SIG_ME, bytes(pk_1), b32_1], [AGG_SIG_UNSAFE, bytes(pk_2), b32_2]]
    )
    pairs = list(verify_pairs_for_conditions(conditions, b32_3))
    msg1 = b32_1 + b32_3
    msg2 = b32_2
    assert pairs == [(pk_1, msg1), (pk_2, msg2)]


def test_hsmk_dump_us():
    # for some reason, if this import is move to the global scope,
    # the `test_cmd` test fails. Maybe the stdout capture fails?
    from hsmk.cmds import hsm_dump_us

    where = pathlib.Path(__file__)
    d = (where.parent / "cmds" / "hsm_dump_us_1.txt").read_text()
    hex_blob = d.split()[1]
    with tempfile.NamedTemporaryFile("w") as f:
        f.write(hex_blob)
        f.flush()
        hsm_dump_us.main([f.name])



from hashlib import sha256

import argparse


from chik_base.bls12_381 import BLSPublicKey
from chik_base.core import Coin, CoinSpend

from hsmk.core.unsigned_spend import UnsignedSpend
from hsmk.puzzles.p2_delegated_puzzle_or_hidden_puzzle import (
    puzzle_for_synthetic_public_key,
    solution_for_conditions,
)
from hsmk.util.byte_chunks import chunks_for_zlib_blob
from hsmk.util.qrint_encoding import b2a_qrint


DEFAULT_PARENT_COIN_ID = sha256(b"fake_id").digest()


def bytes32_fromhex(s: str) -> bytes:
    try:
        b = bytes.fromhex(s)
        if len(b) == 32:
            return b
    except Exception:
        pass
    raise ValueError("not 64 bytes of hex")


DESCRIPTION = (
    "Proof of secret exponent request generator.\n\n"
    'Generate a fake "challenge coin" for the hsmk to sign to prove ownership '
    "of the private key corresponding to a given public key."
)
EPILOG = "Output is in qrint form."


def main():
    parser = argparse.ArgumentParser(description=DESCRIPTION, epilog=EPILOG)
    parser.add_argument(
        "--parent-coin-id",
        type=bytes32_fromhex,
        default=DEFAULT_PARENT_COIN_ID,
        help="64 character hex representing parent coin id",
    )
    parser.add_argument(
        "--network-id",
        type=bytes.fromhex,
        default=b"",
        help="hex representing network id",
    )
    parser.add_argument(
        "-c",
        "--chunk-size",
        type=int,
        default=255,
        help="maximum byte count for each QR code",
    )
    parser.add_argument("-q", "--quiet", action="store_true", help="quiet mode")
    parser.add_argument("bech32m_public_key", help="bech32m-encoded public key")
    parser.add_argument("message", help="message to embed in challenge")
    args = parser.parse_args()

    verbose = not args.quiet

    public_key = BLSPublicKey.from_bech32m(args.bech32m_public_key)
    puzzle = puzzle_for_synthetic_public_key(public_key)
    puzzle_hash = puzzle.tree_hash()

    coin = Coin(args.parent_coin_id, puzzle_hash, 1)
    coin_spend = CoinSpend(coin, puzzle, solution_for_conditions(args.message))
    sum_hints = []
    path_hints = []
    agg_sig_me_network_suffix = b""  # hack to keep unsigned spend shorter
    unsigned_spend = UnsignedSpend(
        [coin_spend], sum_hints, path_hints, agg_sig_me_network_suffix
    )

    with open("us.qr", "w") as f:
        f.write(b2a_qrint(bytes(unsigned_spend)))

    print(f"challenge coin id: {coin.name().hex()}\n")

    blob = bytes(unsigned_spend)
    chunks = [b2a_qrint(_) for _ in chunks_for_zlib_blob(blob, args.chunk_size)]
    if verbose:
        print(f"chunk count: {len(chunks)}\n")

    for chunk in chunks:
        print(chunk)
        if verbose:
            print()

    if verbose:
        print(
            "Output is in qrint form. Use \n\n`segno --compact CHUNK`"
            "\n\nto create a QR code, where `CHUNK` is a qrint value above"
        )
        print(
            "Note: `segno` can also generate PNG files. Use "
            "`segno -h` for more info."
        )


if __name__ == "__main__":
    main()



import argparse

from chik_base.cbincode import from_bytes
from chik_base.core import SpendBundle

from hsmk.debug.debug_spend_bundle import debug_spend_bundle


def file_or_string(p) -> str:
    try:
        with open(p) as f:
            text = f.read().strip()
    except Exception:
        text = p
    return text


def hsmk_dump_sb(args, parser):
    blob = bytes.fromhex(file_or_string(args.spend_bundle))
    spend_bundle = from_bytes(SpendBundle, blob)
    validates = debug_spend_bundle(spend_bundle)
    assert validates is True


def create_parser():
    parser = argparse.ArgumentParser(description="Dump information about `SpendBundle`")
    parser.add_argument(
        "spend_bundle",
        metavar="hex-encoded-spend-bundle-or-file",
        help="hex-encoded `SpendBundle`",
    )
    return parser


def main():
    parser = create_parser()
    args = parser.parse_args()
    return hsmk_dump_sb(args, parser)


if __name__ == "__main__":
    main()



import argparse
import os.path

from hsmk.util.qrint_encoding import a2b_qrint, b2a_qrint


def file_or_string(p) -> str:
    try:
        with open(p) as f:
            text = f.read().strip()
    except Exception:
        text = p
    return text


def decode(path, hex_output):
    if path.endswith(".qri") and len(path) > 4:
        new_path = path[:-4]
    else:
        new_path = path + (".hex" if hex_output else ".bin")
    if os.path.exists(new_path):
        raise ValueError(f"{new_path} already exists")

    blob = open(path).read().strip()
    decoded = a2b_qrint(blob)

    if hex_output:
        with open(new_path, "w") as f:
            f.write(decoded.hex())
        return

    with open(new_path, "wb") as f:
        f.write(decoded)


def encode(path):
    new_path = path + ".qri"
    if os.path.exists(new_path):
        raise ValueError(f"{new_path} already exists")

    blob = open(path, "rb").read()
    encoded = b2a_qrint(blob)
    with open(new_path, "w") as f:
        f.write(encoded)


def qrint(args, parser):
    with open(args.path, "rb") as f:
        blob = f.read()

    is_qrint = (
        all(c in b"0123456789" for c in blob.strip()) and not args.encode_to_qrint
    )

    if is_qrint:
        return decode(args.path, args.hex_output)
    else:
        return encode(args.path)


def create_parser():
    parser = argparse.ArgumentParser(description="Convert binary to/from qrint format.")
    parser.add_argument(
        "-e",
        "--encode-to-qrint",
        action="store_true",
        help="force conversion from binary to qrint",
    )
    parser.add_argument(
        "-H", "--hex-output", action="store_true", help="force convert to hex"
    )
    parser.add_argument(
        "path",
        metavar="path-to-binary-or-qrint-file",
        help="file containing qrint or binary (context-sensitive)",
    )
    return parser


def main():
    parser = create_parser()
    args = parser.parse_args()
    return qrint(args, parser)


if __name__ == "__main__":
    main()



from dataclasses import dataclass

from types import GenericAlias
from typing import (
    Callable,
    get_origin,
    get_args,
    Optional,
    Type,
    TypeVar,
    Generic,
)


Gtype = type | GenericAlias
T = TypeVar("T")
SimpleTypeLookup = dict[Type, T]
CompoundLookup = dict[Type, Callable[[Type, tuple[Type, ...], "TypeTree"], T]]
OtherHandler = Callable[[Type, "TypeTree"], Optional[T]]


@dataclass
class TypeTree(Generic[T]):
    """
    `simple_type_lookup`: a type to callable look-up. Must return a `T` value.
    `compound_type_lookup`: recursively handle compound types like `list` and `tuple`.
    `other_f`: a function to take a type and return a `T` value
    """

    simple_type_lookup: SimpleTypeLookup[T]
    compound_lookup: CompoundLookup[T]
    other_handler: OtherHandler[T]

    def __call__(self, t: Type) -> T:
        """
        Recursively descend a "type tree" invoking the appropriate functions.

        This function is helpful for run-time building a complex function that operates
        on a complex type out of simpler functions that operate on base types.
        """
        origin: None | Type = get_origin(t)
        if origin is not None:
            f = self.compound_lookup.get(origin)
            if f:
                args: tuple[Type, ...] = get_args(t)
                return f(origin, args, self)
        g = self.simple_type_lookup.get(t)
        if g:
            return g
        r = self.other_handler(t, self)
        if r:
            return r
        raise ValueError(f"unable to handle type {t}")



from decimal import Decimal
from typing import BinaryIO, Iterable, List, TextIO

import argparse
import io
import readline  # noqa: F401  this allows long lines on stdin
import subprocess
import sys
import zlib

from chik_base.atoms import bytes32
from chik_base.bls12_381 import BLSSecretExponent, BLSSignature
from chik_base.util.bech32 import bech32_encode


import segno

from hsmk.consensus.conditions import conditions_by_opcode
from hsmk.core.unsigned_spend import UnsignedSpend
from hsmk.process.sign import conditions_for_coin_spend, sign
from hsmk.puzzles import conlang
from hsmk.util.byte_chunks import ChunkAssembler
from hsmk.util.qrint_encoding import a2b_qrint, b2a_qrint


XCK_PER_MOJO = Decimal("1e12")


def unsigned_spend_from_blob(blob: bytes) -> UnsignedSpend:
    try:
        uncompressed_blob = zlib.decompress(blob)
        return UnsignedSpend.from_bytes(uncompressed_blob)
    except Exception:
        return UnsignedSpend.from_bytes(blob)


def create_unsigned_spend_pipeline(
    nochunks: bool, f=sys.stdout
) -> Iterable[UnsignedSpend]:
    print("waiting for qrint-encoded signing requests", file=f)
    partial_encodings = {}
    while True:
        try:
            print("> ", end="", file=f)
            line = input("").strip()
            if len(line) == 0:
                break
            blob = a2b_qrint(line)

            if nochunks:
                yield unsigned_spend_from_blob(blob)
                break

            part_count = blob[-1]
            if part_count not in partial_encodings:
                partial_encodings[part_count] = ChunkAssembler()
            ca = partial_encodings[part_count]
            ca.add_chunk(blob)
            if ca.is_assembled():
                del partial_encodings[part_count]
                blob = ca.assemble()
                yield unsigned_spend_from_blob(blob)
        except EOFError:
            break
        except Exception as ex:
            print(ex, file=f)


def replace_with_gpg_pipe(args, f: BinaryIO) -> TextIO:
    gpg_args = ["gpg", "-d"]
    if args.gpg_argument:
        gpg_args.extend(args.gpg_argument.split())
    gpg_args.append(f.name)
    popen = subprocess.Popen(gpg_args, stdout=subprocess.PIPE)
    if popen is None or popen.stdout is None:
        raise ValueError("couldn't launch gpg")
    return io.TextIOWrapper(popen.stdout)


def parse_private_key_file(args) -> List[BLSSecretExponent]:
    secret_exponents = []
    for f in args.private_key_file:
        if f.name.endswith(".gpg"):
            f = replace_with_gpg_pipe(args, f)
        for line in f.readlines():
            try:
                secret_exponent = BLSSecretExponent.from_bech32m(line.strip())
                secret_exponents.append(secret_exponent)
            except ValueError:
                pass
    return secret_exponents


def summarize_unsigned_spend(unsigned_spend: UnsignedSpend, f=sys.stdout):
    print(file=f)
    for coin_spend in unsigned_spend.coin_spends:
        xck_amount = Decimal(coin_spend.coin.amount) / XCK_PER_MOJO
        address = address_for_puzzle_hash(coin_spend.coin.puzzle_hash)
        print(f"COIN SPENT: {xck_amount:0.12f} xck at address {address}", file=f)
        conditions = conditions_for_coin_spend(coin_spend)

    print(file=f)
    for coin_spend in unsigned_spend.coin_spends:
        conditions = conditions_for_coin_spend(coin_spend)
        conditions_lookup = conditions_by_opcode(conditions)
        for create_coin in conditions_lookup.get(conlang.CREATE_COIN, []):
            puzzle_hash = create_coin.at("rf").atom
            address = address_for_puzzle_hash(puzzle_hash)
            amount = int(create_coin.at("rrf"))
            xck_amount = Decimal(amount) / XCK_PER_MOJO
            print(f"COIN CREATED: {xck_amount:0.12f} xck to {address}", file=f)
    print(file=f)


def address_for_puzzle_hash(puzzle_hash: bytes32) -> str:
    return bech32_encode("xck", puzzle_hash)


def check_ok():
    text = input('if this looks reasonable, enter "ok" to generate signature> ')
    return text.lower() == "ok"


def hsmk(args, parser):
    wallet = parse_private_key_file(args)
    f = sys.stderr
    unsigned_spend_pipeline = create_unsigned_spend_pipeline(args.nochunks, f)
    for unsigned_spend in unsigned_spend_pipeline:
        if not args.yes:
            summarize_unsigned_spend(unsigned_spend, f)
            if not check_ok():
                continue
        signature_info = sign(unsigned_spend, wallet)
        if signature_info:
            signature = sum(
                [_.signature for _ in signature_info], start=BLSSignature.zero()
            )
            encoded_sig = b2a_qrint(bytes(signature))
            if args.qr:
                qr = segno.make_qr(encoded_sig)
                print()
                qr.terminal(compact=True)
                print()
            else:
                print(encoded_sig)


def create_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        description="Manage private keys and process signing requests"
    )
    parser.add_argument(
        "-y",
        "--yes",
        help="skip confirmations",
        action="store_true",
    )
    parser.add_argument(
        "--qr",
        help="show signature as QR code",
        action="store_true",
    )
    parser.add_argument(
        "--nochunks",
        help="read the spend in its entirety rather than as chunks (testing only)",
        action="store_true",
    )
    parser.add_argument(
        "-g", "--gpg-argument", help="argument to pass to gpg (besides -d).", default=""
    )
    parser.add_argument(
        # "-f",
        "private_key_file",
        metavar="path-to-private-keys",
        action="append",
        default=[],
        help=(
            "file containing bech32m-encoded secret exponents. "
            """If file name ends with .gpg, "gpg -d" will be invoked """
            "automatically. File is read one line at a time."
        ),
        type=argparse.FileType("r"),
    )
    return parser


def main(argv=sys.argv[1:]):
    parser = create_parser()
    args = parser.parse_args(argv)
    return hsmk(args, parser)


if __name__ == "__main__":
    main()



from dataclasses import is_dataclass, fields, MISSING
from typing import Any, Callable, Optional, Tuple, Type, Union, get_type_hints

from chik_base.meta.type_tree import ArgsType, CompoundLookup, OriginArgsType, TypeTree
from chik_base.meta.typing import GenericAlias

from klvm_rs import Program  # type: ignore

ToProgram = Callable[[Any], Program]
FromProgram = Callable[[Program], Any]


class EncodingError(ValueError):
    pass


class tuple_frugal(tuple):
    pass


class Frugal:
    """
    This is a tag. Subclasses, when serialized, don't use a nil terminator when
    serialized.
    """

    pass


# some helper methods to implement chik serialization
#
def read_bytes(p: Program) -> bytes:
    if p.atom is None:
        raise EncodingError("expected atom")
    return p.atom


def read_str(p: Program) -> str:
    return read_bytes(p).decode()


def read_int(p: Program) -> int:
    return Program.int_from_bytes(read_bytes(p))


def serialize_for_list(origin, args, type_tree: TypeTree) -> Program:
    write_item = type_tree(args[0])

    def serialize_list(items):
        return Program.to([write_item(_) for _ in items])

    return serialize_list


def serialize_for_optional(origin, args, type_tree: TypeTree) -> Program:
    if len(args) == 2 and type(None) is args[1]:
        write_item = type_tree(args[0])

        def serialize_optional(item):
            if item is None:
                return Program.to((Program.null(), Program.null()))
            else:
                return Program.to((1, write_item(item)))

        return serialize_optional
    else:
        raise ValueError("No serialization support for Union types (besides Optional)")


def serialize_for_tuple(origin, args, type_tree: TypeTree) -> Program:
    write_items = [type_tree(_) for _ in args]

    def serialize_tuple(items):
        item_list = list(items)
        if len(item_list) != len(write_items):
            raise EncodingError("incorrect number of items in tuple")
        return Program.to(
            [
                write_f(item)
                for write_f, item in zip(
                    write_items,
                    items,
                )
            ]
        )

    return serialize_tuple


def ser_for_tuple_frugal(origin, args, type_tree: TypeTree) -> Program:
    streaming_calls = [
        type_tree(
            _,
        )
        for _ in args
    ]

    def ser(item):
        if len(item) != len(streaming_calls):
            raise EncodingError("incorrect number of items in tuple")

        values = list(zip(streaming_calls, item))
        sc, v = values.pop()
        t = sc(v)
        while values:
            sc, v = values.pop()
            t = (sc(v), t)
        return Program.to(t)

    return ser


SERIALIZER_COMPOUND_TYPE_LOOKUP: CompoundLookup[ToProgram] = {
    list: serialize_for_list,
    tuple: serialize_for_tuple,
    tuple_frugal: ser_for_tuple_frugal,
    Union: serialize_for_optional,
}


def types_for_fields(t: type, call_morpher, type_tree: TypeTree):
    # split into key-based and location-based

    key_based = []
    location_based = []
    type_hints = get_type_hints(t)
    for f in fields(t):
        type_hint = type_hints[f.name]
        default_value = (
            f.default if f.default_factory is MISSING else f.default_factory()
        )
        m = f.metadata
        key = m.get("key")
        if key is None:
            location_based.append((f.name, type_hint))
        else:
            alt_serde_type = m.get("alt_serde_type")
            storage_type = alt_serde_type[0] if alt_serde_type else type_hint
            call = type_tree(storage_type)
            key_based.append((key, f.name, call_morpher(call, f), default_value))

    return location_based, key_based


def ser_dataclass(origin: Type, args_type: ArgsType, type_tree: TypeTree) -> Program:
    def morph_call(call, f):
        alt_serde_type = f.metadata.get("alt_serde_type")
        if alt_serde_type:
            _type, from_storage, _to_storage = alt_serde_type

            def f(x):
                return call(from_storage(x))

            return f
        return call

    location_based, key_based = types_for_fields(origin, morph_call, type_tree)

    types = tuple(type_hint for name, type_hint in location_based)
    tuple_type = GenericAlias(tuple, types)
    if key_based:
        types = types + (
            GenericAlias(list, (GenericAlias(tuple_frugal, (str, Program)),)),
        )
    if key_based or issubclass(origin, Frugal):
        tuple_type = GenericAlias(tuple_frugal, types)

    names = tuple(name for name, type_hint in location_based)

    ser_tuple = type_tree(tuple_type)

    def ser(item):
        # convert to a tuple
        v = []
        for name in names:
            v.append(getattr(item, name))
        if key_based:
            d = []
            for key, name, call, default_value in key_based:
                a = getattr(item, name)
                if a == default_value:
                    continue
                d.append((key, call(a)))
            v.append(d)

        return ser_tuple(v)

    return ser


def fail_ser(
    origin: Type, args_type: ArgsType, type_tree: TypeTree
) -> Optional[ToProgram]:
    if issubclass(origin, (str, bytes, int)):
        return Program.to

    if is_dataclass(origin):
        return ser_dataclass(origin, args_type, type_tree)

    if hasattr(origin, "__bytes__"):
        return lambda x: Program.to(bytes(x))

    return None


def deser_dataclass(origin: Type, args_type: ArgsType, type_tree: TypeTree):
    def morph_call(call, f):
        alt_serde_type = f.metadata.get("alt_serde_type")
        if alt_serde_type:
            _type, _from_storage, to_storage = alt_serde_type

            def f(x):
                return to_storage(call(x))

            return f
        return call

    location_based, key_based = types_for_fields(origin, morph_call, type_tree)

    types = tuple(type_hint for name, type_hint in location_based)
    tuple_type = GenericAlias(tuple, types)
    if key_based:
        types = types + (
            GenericAlias(list, GenericAlias(tuple_frugal, (str, Program))),
        )
    if key_based or issubclass(origin, Frugal):
        tuple_type = GenericAlias(tuple_frugal, types)

    de_tuple = type_tree(tuple_type)

    if key_based:

        def de(p: Program):
            the_tuple = de_tuple(p)
            args = the_tuple[:-1]
            d = dict((k, v) for k, v in the_tuple[-1])
            kwargs = {}
            for key, name, call, default_value in key_based:
                if key in d:
                    kwargs[name] = call(d[key])
                else:
                    if default_value == MISSING:
                        raise EncodingError(
                            f"missing required field for {name} with key {key}"
                        )
                    kwargs[name] = default_value

            return origin(*args, **kwargs)

    else:

        def de(p: Program):
            the_tuple = de_tuple(p)
            return origin(*the_tuple)

    return de


def fail_deser(origin: Type, args_type: ArgsType, type_tree: TypeTree):
    if issubclass(origin, int):
        return read_int

    if issubclass(origin, bytes):
        return read_bytes

    if issubclass(origin, str):
        return read_str

    if is_dataclass(origin):
        return deser_dataclass(origin, args_type, type_tree)

    if hasattr(origin, "from_bytes"):
        return lambda p: origin.from_bytes(read_bytes(p))

    return None


def to_program_for_type(t: type) -> Callable[[Any], Program]:
    return TypeTree(
        {(Program, None): lambda x: x},
        SERIALIZER_COMPOUND_TYPE_LOOKUP,
        fail_ser,
    )(t)


def deser_for_list(origin, args, type_tree: TypeTree):
    read_item = type_tree(args[0])

    def deserialize_list(p: Program) -> list:
        return [read_item(_) for _ in p.as_iter()]

    return deserialize_list


def deser_for_tuple(origin, args, type_tree: TypeTree):
    read_items = [type_tree(_) for _ in args]

    def deserialize_tuple(p: Program) -> Tuple[Any, ...]:
        items = list(p.as_iter())
        if len(items) != len(read_items):
            raise EncodingError("wrong size program")
        return tuple(f(_) for f, _ in zip(read_items, items))

    return deserialize_tuple


def de_for_tuple_frugal(origin, args, type_tree: TypeTree):
    read_items = [type_tree(_) for _ in args]

    def de(p: Program) -> Tuple[Any, ...]:
        args = []
        todo = list(reversed(read_items))
        while todo:
            des = todo.pop()
            if todo:
                v = p.pair[0]
                p = p.pair[1]
            else:
                v = p
            args.append(des(Program.to(v)))
        return tuple(args)

    return de

def deser_for_optional(origin, args, type_tree: TypeTree):
    if len(args) == 2 and type(None) is args[1]:
        read_item = type_tree(args[0])

        def deserialize_optional(p: Program):
            if p.first() == Program.null():
                return None
            else:
                return read_item(p.rest())

        return deserialize_optional
    else:
        raise ValueError("No serialization support for Union types (besides Optional)")


DESERIALIZER_COMPOUND_TYPE_LOOKUP: CompoundLookup[FromProgram] = {
    list: deser_for_list,
    tuple: deser_for_tuple,
    tuple_frugal: de_for_tuple_frugal,
    Union: deser_for_optional,
}


def from_program_for_type(t: type) -> FromProgram:
    simple_lookup: dict[OriginArgsType, FromProgram] = {
        (Program, None): lambda x: x,
    }
    return TypeTree(
        simple_lookup,
        DESERIALIZER_COMPOUND_TYPE_LOOKUP,
        fail_deser,
    )(t)



from dataclasses import dataclass
from typing import List

from chik_base.atoms import bytes32
from chik_base.bls12_381 import BLSPublicKey, BLSSignature
from chik_base.core import Coin, CoinSpend

from klvm_rs import Program

from .klvm_serialization import (
    as_atom,
    as_int,
    klvm_to_list,
    no_op,
    transform_dict,
    transform_dict_by_key,
    transform_as_struct,
)
from .signing_hints import SumHint, PathHint


@dataclass
class SignatureInfo:
    signature: BLSSignature
    partial_public_key: BLSPublicKey
    final_public_key: BLSPublicKey
    message: bytes


@dataclass
class UnsignedSpend:
    coin_spends: List[CoinSpend]
    sum_hints: List[SumHint]
    path_hints: List[PathHint]
    agg_sig_me_network_suffix: bytes32

    def as_program(self):
        as_klvm = [("a", self.agg_sig_me_network_suffix)]
        cs_as_klvm = [
            [_.coin.parent_coin_info, _.puzzle_reveal, _.coin.amount, _.solution]
            for _ in self.coin_spends
        ]
        as_klvm.append(("c", cs_as_klvm))
        sh_as_klvm = [_.as_program() for _ in self.sum_hints]
        as_klvm.append(("s", sh_as_klvm))
        ph_as_klvm = [_.as_program() for _ in self.path_hints]
        as_klvm.append(("p", ph_as_klvm))
        self_as_program = Program.to(as_klvm)
        return self_as_program

    @classmethod
    def from_program(cls, program) -> "UnsignedSpend":
        d = transform_dict(program, transform_dict_by_key(UNSIGNED_SPEND_TRANSFORMER))
        return cls(d["c"], d.get("s", []), d.get("p", []), d["a"])


def coin_spend_from_program(program: Program) -> CoinSpend:
    struct = transform_as_struct(program, as_atom, no_op, as_int, no_op)
    parent_coin_info, puzzle_reveal, amount, solution = struct
    return CoinSpend(
        Coin(
            parent_coin_info,
            puzzle_reveal.tree_hash(),
            amount,
        ),
        puzzle_reveal,
        solution,
    )


UNSIGNED_SPEND_TRANSFORMER = {
    "c": lambda x: klvm_to_list(x, coin_spend_from_program),
    "s": lambda x: klvm_to_list(x, SumHint.from_program),
    "p": lambda x: klvm_to_list(x, PathHint.from_program),
    "a": lambda x: x.atom,
}



import secrets

from chik_base.bls12_381 import BLSSecretExponent


def main():
    secret_exponent = BLSSecretExponent.from_int(secrets.randbits(256))
    b = secret_exponent.as_bech32m()
    assert BLSSecretExponent.from_bech32m(b) == secret_exponent
    print(b)


if __name__ == "__main__":
    main()



from hashlib import sha256

import argparse

from chik_base.bls12_381 import BLSPublicKey, BLSSignature


DEFAULT_PARENT_COIN_ID = sha256(b"fake_id").digest()


def bytes32_fromhex(s: str) -> bytes:
    try:
        b = bytes.fromhex(s)
        if len(b) == 32:
            return b
    except Exception:
        pass
    raise ValueError("not 64 bytes of hex")


DESCRIPTION = "Proof of secret exponent request verifier."
EPILOG = (
    'Note: signature output from hsmk is in "qrint" form. '
    "Use `qrint -H` to convert it to hex."
)


def create_parser():
    parser = argparse.ArgumentParser(description=DESCRIPTION, epilog=EPILOG)
    parser.add_argument(
        "--network-id",
        type=bytes.fromhex,
        default=b"",
        help="hex representing network id",
    )
    parser.add_argument("bech32m_public_key", help="bech32m-encoded public key")
    parser.add_argument("message", help="message to embed in challenge")
    parser.add_argument(
        "challenge_coin_id",
        type=bytes32_fromhex,
        help="64 character hex representing coin id generated by challenge",
    )
    parser.add_argument(
        "signature",
        type=bytes.fromhex,
        help="hex representing signature response to challenge",
    )
    return parser


def main():
    parser = create_parser()
    args = parser.parse_args()

    ONE = bytes([1])
    TWO = bytes([2])

    coin_name = args.challenge_coin_id
    pk = BLSPublicKey.from_bech32m(args.bech32m_public_key)

    try:
        sig = BLSSignature.from_bytes(args.signature)
    except Exception:
        pass

    def s256(b):
        return sha256(b).digest()

    hashed_message = s256(TWO + s256(ONE + ONE) + s256(ONE + args.message.encode()))

    signed_message = hashed_message + coin_name

    r = sig.verify([(pk, signed_message)])

    if r:
        print("message signed correctly")
    else:
        print("BAD SIGNATURE")


if __name__ == "__main__":
    main()



from hsmk.util.qrint_encoding import a2b_qrint, b2a_qrint, b2a_qrint_payload

from .generate import bytes32_generate

MAX_SIZE = 8192
BIG_BLOB = bytes([bytes32_generate(_)[0] for _ in range(MAX_SIZE)])


def check_b2a(data, expected_value=None):
    b = b2a_qrint(data)
    d = a2b_qrint(b)
    assert data == d
    if expected_value is not None:
        assert b == expected_value


def test_b2a_qrint():
    check_b2a(b"", "2")
    check_b2a(bytes([1]), "1002")
    check_b2a(bytes([2]), "1004")
    check_b2a(bytes([200]), "1620")
    check_b2a(bytes([1, 200]), "1003440")
    check_b2a(bytes([187, 200]), "1567440")
    check_b2a(
        b"the quick brown fox jumps over the lazy dogs",
        "33905997376761408039757883411394151763462345499342126400"
        "1026867145388026209696872339324742682190340016362470400",
    )
    RANGE = list(range(0, 800)) + list(range(800, MAX_SIZE, 5))
    for size in RANGE:
        start = (MAX_SIZE - size) // 2
        blob = BIG_BLOB[start : start + size]
        check_b2a(blob)



AGG_SIG_UNSAFE = 49
AGG_SIG_ME = 50

# the conditions below reserve coin amounts and
# have to be accounted for in output totals

CREATE_COIN = 51
RESERVE_FEE = 52

# the conditions below deal with announcements, for inter-coin communication

# coin announcements
CREATE_COIN_ANNOUNCEMENT = 60
ASSERT_COIN_ANNOUNCEMENT = 61

# puzzle announcements
CREATE_PUZZLE_ANNOUNCEMENT = 62
ASSERT_PUZZLE_ANNOUNCEMENT = 63

# the conditions below let coins inquire about themselves

ASSERT_MY_COIN_ID = 70
ASSERT_MY_PARENT_ID = 71
ASSERT_MY_PUZZLEHASH = 72
ASSERT_MY_AMOUNT = 73

# the conditions below ensure that we're "far enough" in the future

# wall-clock time
ASSERT_SECONDS_RELATIVE = 80
ASSERT_SECONDS_ABSOLUTE = 81

# block index
ASSERT_HEIGHT_RELATIVE = 82
ASSERT_HEIGHT_ABSOLUTE = 83



import zlib

from tests.generate import se_generate, bytes32_generate, uint256_generate

from chik_base.core import Coin, CoinSpend, SpendBundle
from chik_base.cbincode.util import from_bytes, to_bytes

from hsmk.core.signing_hints import SumHint, PathHint
from hsmk.core.unsigned_spend import UnsignedSpend
from hsmk.debug.debug_spend_bundle import debug_spend_bundle
from hsmk.puzzles.p2_delegated_puzzle_or_hidden_puzzle import (
    DEFAULT_HIDDEN_PUZZLE,
    puzzle_for_public_key_and_hidden_puzzle,
    solution_for_conditions,
    calculate_synthetic_offset,
)
from hsmk.process.sign import sign, generate_synthetic_offset_signatures
from hsmk.puzzles.conlang import CREATE_COIN
from hsmk.util.byte_chunks import (
    ChunkAssembler,
    chunks_for_zlib_blob,
    create_chunks_for_blob,
)


AGG_SIG_ME_ADDITIONAL_DATA = bytes.fromhex(
    "ccd5bb71183532bff220ba46c268991a3ff07eb358e8255a65c30a2dce0e5fbb"
)


DEFAULT_HIDDEN_PUZZLE_HASH = DEFAULT_HIDDEN_PUZZLE.tree_hash()


def test_lifecycle():
    """
    let's create three `CoinSpend` objects and generate a signing request
    then pass it to the hsmk0 tool for signatures
    then reassemble the signatures
    """

    # create two private keys A and B

    se_A = se_generate(100)
    se_B = se_generate(200)

    # create some compound public keys
    # each using a different path from A and B

    A_PATHS = [[1, 5, 10], [2, 3, 1], [1000, 0, 92]]
    B_PATHS = [[1, 5, 10], [23, 33, 13], [200, 11, 9992]]

    pks_A = [se_A.child_for_path(_).public_key() for _ in A_PATHS]
    pks_B = [se_B.child_for_path(_).public_key() for _ in B_PATHS]

    # create "sum public keys" that are the sum of pubkeys from each of A and B
    sum_pks = [a + b for a, b in zip(pks_A, pks_B)]

    # create a standard puzzle using the sum public keys
    puzzles = [
        puzzle_for_public_key_and_hidden_puzzle(pk, DEFAULT_HIDDEN_PUZZLE)
        for pk in sum_pks
    ]

    # work out the synthetic private keys as we'll need it later to finish the signing
    synthetic_se_list = [
        calculate_synthetic_offset(pk, DEFAULT_HIDDEN_PUZZLE_HASH) for pk in sum_pks
    ]

    COIN_MAX = 1000000

    coins = [
        Coin(
            bytes32_generate(idx),
            puzzle.tree_hash(),
            5 + (uint256_generate(idx) % COIN_MAX),
        )
        for idx, puzzle in enumerate(puzzles)
    ]

    for coin in coins:
        c = from_bytes(Coin, to_bytes(coin))
        assert c == coin

    # the destination puzzle hashes are nonsense, but that's okay
    dest_puzzle_hashes = [
        bytes32_generate(idx, "dest") for idx, _ in enumerate(puzzles)
    ]

    conditions_for_spends = [
        [[CREATE_COIN, dph, coin.amount]]
        for dph, coin in zip(dest_puzzle_hashes, coins)
    ]
    solutions = [
        solution_for_conditions(conditions) for conditions in conditions_for_spends
    ]
    coin_spends = [CoinSpend(*triple) for triple in zip(coins, puzzles, solutions)]

    # create the "sum hints" which give information on how the public keys that
    # appear in the `AGG_SIG_ME` conditions were constructed

    sum_hints = [
        SumHint([a_pk, b_pk], synthetic_se)
        for a_pk, b_pk, synthetic_se in zip(pks_A, pks_B, synthetic_se_list)
    ]

    # create the "path hints" which give information on how to get from the
    # root public key to the public keys used as inputs to the sum keys
    #
    # If the public key is also the root public key, you don't need this hint

    pk_A = se_A.public_key()
    path_hints_a = [PathHint(pk_A, path) for pk, path in zip(pks_A, A_PATHS)]
    pk_B = se_B.public_key()
    path_hints_b = [PathHint(pk_B, path) for pk, path in zip(pks_B, B_PATHS)]

    path_hints = path_hints_a + path_hints_b

    print(sum_hints)
    print(path_hints)

    unsigned_spend = UnsignedSpend(
        coin_spends, sum_hints, path_hints, AGG_SIG_ME_ADDITIONAL_DATA
    )

    chunks = chunks_for_zlib_blob(bytes(unsigned_spend), 500)

    assert unsigned_spend == UnsignedSpend.from_bytes(
        zlib.decompress(ChunkAssembler(chunks).assemble())
    )

    spend_chunks = create_chunks_for_blob(bytes(unsigned_spend), 250)
    assembler = ChunkAssembler()
    assembler.add_chunk(spend_chunks[0])
    assembler.add_chunk(spend_chunks[0])
    assert assembler.status() == (1, len(spend_chunks))
    assert not assembler.is_assembled()
    for i in range(1, len(spend_chunks)):
        assembler.add_chunk(spend_chunks[i])
    assert assembler.is_assembled()
    assert assembler.status() == (len(spend_chunks), len(spend_chunks))
    assert unsigned_spend == UnsignedSpend.from_bytes(assembler.assemble())

    print("-" * 10)

    # this simulates giving the `UnsignedSpend` to A and getting signatures back

    signatures_A = sign(unsigned_spend, [se_A])
    print(signatures_A)

    # this simulates giving the `UnsignedSpend` to B and getting signatures back

    signatures_B = sign(unsigned_spend, [se_B])
    print(signatures_B)

    signatures = signatures_A + signatures_B
    spend_bundle = create_spend_bundle(unsigned_spend, signatures)

    sb2 = from_bytes(SpendBundle, to_bytes(spend_bundle))
    assert sb2 == spend_bundle

    validates = debug_spend_bundle(spend_bundle)
    assert validates is True


def create_spend_bundle(unsigned_spend, signatures):
    extra_signatures = generate_synthetic_offset_signatures(unsigned_spend)

    # now let's try adding them all together and creating a `SpendBundle`

    all_signatures = [sig_info.signature for sig_info in signatures + extra_signatures]
    total_signature = sum(all_signatures, start=all_signatures[0].zero())

    return SpendBundle(unsigned_spend.coin_spends, total_signature)



import argparse
import hashlib
import sys
import zlib

from klvm_rs import Program  # type: ignore

from chik_base.bls12_381 import BLSPublicKey
from chik_base.core import Coin, CoinSpend

from hsmk.core.signing_hints import SumHint, PathHint
from hsmk.core.unsigned_spend import UnsignedSpend
from hsmk.puzzles.p2_delegated_puzzle_or_hidden_puzzle import (
    DEFAULT_HIDDEN_PUZZLE,
    puzzle_for_public_key_and_hidden_puzzle,
    solution_for_conditions,
    calculate_synthetic_offset,
)
from hsmk.puzzles.conlang import CREATE_COIN
from hsmk.util.byte_chunks import (
    create_chunks_for_blob,
    optimal_chunk_size_for_max_chunk_size,
)
from hsmk.util.qrint_encoding import b2a_qrint

MAINNET_AGG_SIG_ME_ADDITIONAL_DATA = bytes.fromhex(
    "ccd5bb71183532bff220ba46c268991a3ff07eb358e8255a65c30a2dce0e5fbb"
)


DEFAULT_HIDDEN_PUZZLE_HASH = DEFAULT_HIDDEN_PUZZLE.tree_hash()


def hsm_test_spend(args, parser):
    root_public_keys = [BLSPublicKey.from_bech32m(_) for _ in args.public_key]

    paths = [[index, index + 1] for index in range(len(root_public_keys))]

    public_keys = [
        root_key.child_for_path(path) for root_key, path in zip(root_public_keys, paths)
    ]

    # create "sum public keys" that are the sum of pubkeys from each of A and B
    sum_pk = sum(public_keys, start=BLSPublicKey.zero())

    # create a standard puzzle using the sum of the public keys
    puzzle = puzzle_for_public_key_and_hidden_puzzle(sum_pk, DEFAULT_HIDDEN_PUZZLE)

    # make the coin
    FAKE_PARENT = hashlib.sha256(b"parent").digest()
    coin = Coin(FAKE_PARENT, puzzle.tree_hash(), 1)

    synthetic_secret_exponent = calculate_synthetic_offset(
        sum_pk, DEFAULT_HIDDEN_PUZZLE_HASH
    )

    sum_hints = [SumHint(public_keys, synthetic_secret_exponent)]

    path_hints = [
        PathHint(root_key, path) for root_key, path in zip(root_public_keys, paths)
    ]

    # destination

    dest_puzzle_hash_1 = Program.to(100).tree_hash()
    dest_puzzle_hash_2 = Program.to(200).tree_hash()
    conditions_for_spend = Program.to(
        [
            [CREATE_COIN, dest_puzzle_hash_1, int(3 * 1e12)],
            [CREATE_COIN, dest_puzzle_hash_2, int(2 * 1e12)],
        ]
    )
    solution = solution_for_conditions(conditions_for_spend)

    coin_spend = CoinSpend(coin, puzzle, solution)

    unsigned_spend = UnsignedSpend(
        [coin_spend], sum_hints, path_hints, MAINNET_AGG_SIG_ME_ADDITIONAL_DATA
    )

    b = bytes(unsigned_spend)
    if args.hex:
        print(b.hex())
    else:
        if args.no_chunks:
            chunks = [b]
        else:
            cb = zlib.compress(b)
            optimal_size = optimal_chunk_size_for_max_chunk_size(
                len(cb), args.max_chunk_size
            )
            chunks = create_chunks_for_blob(cb, optimal_size)
        for chunk in chunks:
            print(b2a_qrint(chunk))

    us = UnsignedSpend.from_bytes(b)
    assert bytes(us) == b


def create_parser():
    parser = argparse.ArgumentParser(
        description="Generate a `UnsignedSpend` test as a proof-of-concept"
    )
    parser.add_argument(
        "-m",
        "--max-chunk-size",
        metavar="maximum-bytes-per-chunk",
        default=8192,
        help="maximum number of bytes encoded into each chunk",
        type=int,
    )
    parser.add_argument(
        "-H",
        "--hex",
        action="store_true",
        help="hex output",
    )
    parser.add_argument(
        "-n",
        "--no-chunks",
        action="store_true",
        help="don't compress or chunk output",
    )
    parser.add_argument(
        "public_key",
        metavar="public-key",
        nargs="+",
        help="bech32m-encoded public key",
        type=str,
    )
    return parser


def main(argv=sys.argv[1:]):
    parser = create_parser()
    args = parser.parse_args(argv)
    return hsm_test_spend(args, parser)


if __name__ == "__main__":  # pragma: no cover
    main()



"""
Pay to conditions

In this puzzle program, the solution is ignored. The reveal of the puzzle
returns a fixed list of conditions. This roughly corresponds to OP_SECURETHEBAG
in bitcoin.

This is a pretty useless most of the time. But some (most?) solutions
require a delegated puzzle program, so in those cases, this is just what
the doctor ordered.
"""

from klvm_rs import Program  # type: ignore

from chiklisp_puzzles import load_puzzle  # type: ignore

MOD = load_puzzle("p2_conditions")


def puzzle_for_conditions(conditions) -> Program:
    return MOD.run_with_cost([conditions], max_cost=1<<32)[1]


def solution_for_conditions(conditions) -> Program:
    return Program.to([puzzle_for_conditions(conditions), 0])



from dataclasses import dataclass
from typing import Dict, List

from chik_base.bls12_381 import BLSPublicKey, BLSSecretExponent

from .klvm_serialization import (
    klvm_to_list_of_ints,
    klvm_to_list,
)


@dataclass
class SumHint:
    public_keys: List[BLSPublicKey]
    synthetic_offset: BLSSecretExponent

    def final_public_key(self) -> BLSPublicKey:
        return sum(self.public_keys, start=self.synthetic_offset.public_key())

    def as_program(self):
        return ([bytes(_) for _ in self.public_keys], bytes(self.synthetic_offset))

    @classmethod
    def from_program(cls, program) -> "SumHint":
        public_keys = klvm_to_list(
            program.pair[0], lambda x: BLSPublicKey.from_bytes(x.atom)
        )
        synthetic_offset = BLSSecretExponent.from_bytes(program.pair[1].atom)
        return cls(public_keys, synthetic_offset)


@dataclass
class PathHint:
    root_public_key: BLSPublicKey
    path: List[int]

    def public_key(self) -> BLSPublicKey:
        return self.root_public_key.child_for_path(self.path)

    def as_program(self):
        return (bytes(self.root_public_key), self.path)

    @classmethod
    def from_program(cls, program) -> "PathHint":
        root_public_key = BLSPublicKey.from_bytes(program.pair[0].atom)
        path = klvm_to_list_of_ints(program.pair[1])
        return cls(root_public_key, path)


SumHints = Dict[BLSPublicKey, SumHint]
PathHints = Dict[BLSPublicKey, PathHint]



from typing import Dict, List

from klvm_rs import Program  # type: ignore


def conditions_by_opcode(conditions: Program) -> Dict[int, List[Program]]:
    d: Dict[int, List[Program]] = {}
    for _ in conditions.as_iter():
        if _.pair:
            d.setdefault(Program.to(_.pair[0]).as_int(), []).append(_)
    return d



"""
Pay to delegated puzzle or hidden puzzle

In this puzzle program, the solution must choose either a hidden puzzle or a
delegated puzzle on a given public key.

The given public key is morphed by adding an offset from the hash of the hidden
puzzle and itself, giving a new so-called "synthetic" public key which has the
hidden puzzle hidden inside of it.

If the hidden puzzle path is taken, the hidden puzzle and original public key
will be revealed which proves that it was hidden there in the first place.

This roughly corresponds to bitcoin's taproot.

Note:

p2_delegated_puzzle_or_hidden_puzzle is essentially the "standard coin" in chik.
DEFAULT_HIDDEN_PUZZLE_HASH from this puzzle is used with
calculate_synthetic_secret_key in the wallet's standard pk_to_sk finder.

This is important because it allows sign_coin_spends to function properly via the
following mechanism:

- A 'standard coin' coin exists in the blockchain with some puzzle hash.

- The user's wallet contains a primary sk/pk pair which are used to derive to one
  level a set of auxiliary sk/pk pairs which are used for specific coins. These
  can be used for signing in AGG_SIG_ME, but the standard coin uses a key further
  derived from one of these via calculate_synthetic_secret_key as described in
  https://chiklisp.com/docs/standard_transaction. Therefore, when a wallet needs
  to find a secret key for signing based on a public key, it needs to try repeating
  this derivation as well and see if the BLSPublicKey (pk) associated with any of
  the derived secret keys matches the pk requested by the coin.

- Python code previously appeared which was written like:

    delegated_puzzle_solution = Program.to((1, condition_args))
    solutions = Program.to([[], delegated_puzzle_solution, []])

  In context, delegated_puzzle_solution here is any *chiklisp program*, here one
  simply quoting a list of conditions, and the following argument is the arguments
  to this program, which here are unused. Secondly, the actual arguments to the
  p2_delegated_puzzle_or_hidden_puzzle are given. The first argument determines
  whether a hidden or revealed puzzle is used. If the puzzle is hidden, then what
  is required is a signature given a specific synthetic key since the key cannot be
  derived inline without the puzzle. In that case, the first argument is this key.
  In most cases, the puzzle will be revealed, and this argument will be the nil
  object, () (represented here by an empty python list).

  The second and third arguments are a chiklisp program and its corresponding
  arguments, which will be run inside the standard coin puzzle. This interacts with
  sign_coin_spend in that the AGG_SIG_ME condition added by the inner puzzle asks
  the surrounding system to provide a signature over the provided program with a 
  synthetic key whose derivation is within. Any wallets which intend to use
  standard coins in this way must try to resolve a public key to a secret key
  via this derivation.
"""

import hashlib

from klvm_rs import Program  # type: ignore

from chik_base.atoms import bytes32
from chik_base.bls12_381 import BLSPublicKey, BLSSecretExponent

from chiklisp_puzzles import load_puzzle  # type: ignore

from .p2_conditions import puzzle_for_conditions

DEFAULT_HIDDEN_PUZZLE = Program.from_bytes(
    bytes.fromhex("ff0980")
)  # this puzzle `(=)` always fails

DEFAULT_HIDDEN_PUZZLE_HASH = DEFAULT_HIDDEN_PUZZLE.tree_hash()

MOD = load_puzzle("p2_delegated_puzzle_or_hidden_puzzle")

SYNTHETIC_MOD = load_puzzle("calculate_synthetic_public_key")

MAX_COST = 1 << 24


def calculate_synthetic_offset(
    public_key: BLSPublicKey, hidden_puzzle_hash: bytes32
) -> BLSSecretExponent:
    blob = hashlib.sha256(bytes(public_key) + hidden_puzzle_hash).digest()
    offset = Program.to(blob).as_int()
    assert offset == int(Program.to(blob))
    return BLSSecretExponent.from_int(offset)


def calculate_synthetic_public_key(
    public_key: BLSPublicKey, hidden_puzzle_hash: bytes32
) -> BLSPublicKey:
    _cost, r = SYNTHETIC_MOD.run_with_cost(
        [bytes(public_key), hidden_puzzle_hash], max_cost=MAX_COST
    )
    return BLSPublicKey.from_bytes(r.atom)


def calculate_synthetic_secret_key(
    secret_key: BLSSecretExponent, hidden_puzzle_hash: bytes32
) -> BLSSecretExponent:
    public_key = secret_key.public_key()
    synthetic_offset = calculate_synthetic_offset(public_key, hidden_puzzle_hash)
    synthetic_secret_key = secret_key + synthetic_offset
    return synthetic_secret_key


def puzzle_for_synthetic_public_key(synthetic_public_key: BLSPublicKey) -> Program:
    return MOD.curry(bytes(synthetic_public_key))


def puzzle_for_public_key_and_hidden_puzzle_hash(
    public_key: BLSPublicKey, hidden_puzzle_hash: bytes32
) -> Program:
    synthetic_public_key = calculate_synthetic_public_key(
        public_key, hidden_puzzle_hash
    )
    return puzzle_for_synthetic_public_key(synthetic_public_key)


def puzzle_for_public_key_and_hidden_puzzle(
    public_key: BLSPublicKey, hidden_puzzle: Program
) -> Program:
    return puzzle_for_public_key_and_hidden_puzzle_hash(
        public_key, hidden_puzzle.tree_hash()
    )


def solution_for_delegated_puzzle(
    delegated_puzzle: Program, solution: Program
) -> Program:
    return Program.to([[], delegated_puzzle, solution])


def solution_for_hidden_puzzle(
    hidden_public_key: BLSPublicKey,
    hidden_puzzle: Program,
    solution_to_hidden_puzzle: Program,
) -> Program:
    return Program.to([hidden_public_key, hidden_puzzle, solution_to_hidden_puzzle])


def solution_for_conditions(conditions) -> Program:
    delegated_puzzle = puzzle_for_conditions(conditions)
    return solution_for_delegated_puzzle(delegated_puzzle, Program.to(0))



from pathlib import Path

import argparse
import os
import secrets
import time

import segno

from chik_base.bls12_381 import BLSSecretExponent

import hsmk.cmds.hsmk


def wait_for_usb_mount(parser, mount_path, device_to_mount, mount_command_read_only):
    print(f"insert USB stick corresponding to {device_to_mount}")
    while 1:
        if device_to_mount.exists():
            cmd = mount_command_read_only.format(
                device_to_mount=device_to_mount, mount_path=mount_path
            )
            r = os.system(cmd)
            if r == 0:
                return
            parser.exit(1, f"failed to mount device {device_to_mount} read-only")
        time.sleep(1)


def generate_secret(wallet_path):
    CLEAR_SCREEN = "\033[2J"
    secret_exponent = BLSSecretExponent.from_int(secrets.randbits(256))
    b = secret_exponent.as_bech32m()
    assert BLSSecretExponent.from_bech32m(b) == secret_exponent

    print("No secret found. We are generating you a new secret.")
    while True:
        print(f"write down your secret:\n\n{b}\n")
        input("hit return when done> ")
        print(CLEAR_SCREEN)
        t = input("enter your secret> ")
        if t == b:
            break
        diff_string = "".join(" " if a1 == a2 else "^" for a1, a2 in zip(t, b))
        print(f"{b} <= actual secret")
        print(f"{t} <= you entered")
        print(diff_string)
        print("fix it and let's try again")
        print()

    with open(wallet_path, "w") as f:
        f.write(t)

    print(CLEAR_SCREEN)
    print("you entered your secret correctly! Good job")

    public_key = secret_exponent.public_key().as_bech32m()
    print(f"your public key is {public_key}")
    print("Take a photo of it and share with your coordinator:")
    qr = segno.make_qr(public_key)
    print()
    qr.terminal(compact=True)


def create_parser():
    parser = argparse.ArgumentParser(
        description="Wizard to look for USB mount point and key file and launch hsmk"
    )
    parser.add_argument(
        "-d",
        "--device-to-mount",
        help=(
            "path, usually to something in `/dev/disk/` "
            "to ensure exists before attempting to mount"
        ),
    )
    parser.add_argument(
        "-r",
        "--mount-command-read-only",
        help="command to mount disk read-only",
        default="/usr/bin/sudo mount -o ro {device_to_mount} {mount_path}",
    )
    parser.add_argument(
        "-w",
        "--remount-command-read-write",
        help="command to remount disk read-write",
        default="/usr/bin/sudo mount -o remount,rw {device_to_mount} {mount_path}",
    )
    parser.add_argument(
        "-u",
        "--unmount-command",
        help="command to unmount disk",
        default="/usr/bin/sudo umount {mount_path}",
    )
    parser.add_argument(
        "path_to_secret_exponent_file",
        metavar="bech32m-encoded-se-file",
    )
    return parser


def main():
    parser = create_parser()
    args = parser.parse_args()
    wallet_path = Path(args.path_to_secret_exponent_file)
    mount_path = wallet_path.parent
    if args.device_to_mount:
        wait_for_usb_mount(
            parser, mount_path, Path(args.device_to_mount), args.mount_command_read_only
        )
    if wallet_path.exists():
        parser = hsmk.cmds.hsmk.create_parser()
        args = parser.parse_args(["--qr", str(wallet_path)])
        hsmk.cmds.hsmk.hsmk(args, parser)
    else:
        if args.remount_command_read_write:
            cmd = args.remount_command_read_write.format(
                device_to_mount=args.device_to_mount, mount_path=mount_path
            )
            r = os.system(cmd)
            if r != 0:
                parser.exit(
                    1, f"failed to remount device {args.device_to_mount} read/write"
                )
            cmd = f"/usr/bin/touch {wallet_path}.tmp && /bin/rm {wallet_path}.tmp"
            r = os.system(cmd)
            if r != 0:
                parser.exit(
                    1,
                    f"could not create temporary file `{wallet_path}.tmp`, "
                    f"drive permissions may be wrong",
                )
        generate_secret(wallet_path)
    print()
    input("hit return to power off> ")
    os.system("poweroff")


if __name__ == "__main__":
    main()



import hashlib
import hmac

from chik_base.atoms import bytes32
from chik_base.bls12_381 import BLSPublicKey, BLSSecretExponent


def bytes32_generate(nonce: int, namespace: str = "bytes32") -> bytes32:
    return bytes32(
        hmac.HMAC(
            namespace.encode(), nonce.to_bytes(32, "big"), digestmod=hashlib.sha256
        ).digest()
    )


def uint256_generate(nonce: int, namespace: str = "uint256") -> int:
    return int.from_bytes(bytes32_generate(nonce, namespace), "big")


def se_generate(nonce: int, namespace: str = "BLSSecretExponent") -> BLSSecretExponent:
    return BLSSecretExponent.from_int(uint256_generate(nonce))


def pk_generate(nonce: int, namespace: str = "BLSPublicKey") -> BLSPublicKey:
    return se_generate(nonce, namespace).public_key()



from dataclasses import dataclass, field
from typing import List, Tuple

from chik_base.bls12_381 import BLSPublicKey, BLSSignature
from chik_base.core import Coin, CoinSpend

from klvm_rs import Program  # type: ignore

from hsmk.klvm_serde import (
    to_program_for_type,
    from_program_for_type,
)
from .signing_hints import PathHint, SumHint


# `CoinSpend` objects have the puzzle reveal and the puzzle hash (in the coin)
# which is a little redundant. Casting to `CSTuple` removes the redundant
# puzzle hash, saving 32 bytes when serialized.

CSTuple = Tuple[bytes, Program, int, Program]
SerdeCoinSpends = List[CSTuple]


def to_storage(
    coin_spend_tuples: SerdeCoinSpends,
) -> List[CoinSpend]:
    return [
        CoinSpend(Coin(_[0], _[1].tree_hash(), _[2]), _[1], _[3])
        for _ in coin_spend_tuples
    ]


def from_storage(
    coin_spends: List[CoinSpend],
) -> SerdeCoinSpends:
    return [
        (
            _.coin.parent_coin_info,
            _.puzzle_reveal,
            _.coin.amount,
            _.solution,
        )
        for _ in coin_spends
    ]


@dataclass
class SignatureInfo:
    signature: BLSSignature
    partial_public_key: BLSPublicKey
    final_public_key: BLSPublicKey
    message: bytes


@dataclass
class UnsignedSpend:
    coin_spends: List[CoinSpend] = field(
        metadata=dict(
            key="c",
            alt_serde_type=(
                SerdeCoinSpends,
                from_storage,
                to_storage,
            ),
        ),
    )
    sum_hints: List[SumHint] = field(
        default_factory=list,
        metadata=dict(key="s"),
    )
    path_hints: List[PathHint] = field(
        default_factory=list,
        metadata=dict(key="p"),
    )
    agg_sig_me_network_suffix: bytes = field(
        default=b"",
        metadata=dict(key="a"),
    )

    def __bytes__(self):
        return bytes(TO_PROGRAM(self))

    @classmethod
    def from_bytes(cls, blob: bytes):
        return FROM_PROGRAM(Program.from_bytes(blob))


TO_PROGRAM = to_program_for_type(UnsignedSpend)
FROM_PROGRAM = from_program_for_type(UnsignedSpend)



import io

from klvm_rs import Program  # type: ignore


# this differs from klvm_tools in that it adds the single quote
# and promises to handle it carefully

PRINTABLE = (
    "0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ"
    r"""!"#$%&'()*+,-./:;<=>?@[]^_`{|}~ """
)


def type_for_atom(atom) -> str:
    if len(atom) > 2:
        try:
            v = atom.decode("utf8")
            if all(c in PRINTABLE for c in v):
                if '"' in v:
                    if "'" in v:
                        return "H"
                    return "S"
                return "D"
        except UnicodeDecodeError:
            pass
        return "H"
    if Program.int_to_bytes(Program.int_from_bytes(atom)) == atom:
        return "I"
    return "H"


KWS = (
    # core opcodes 0x01-x08
    ". q a i c f r l x "
    # opcodes on atoms as strings 0x09-0x0f
    "= >s sha256 substr strlen concat . "
    # opcodes on atoms as ints 0x10-0x17
    "+ - * / divmod > ash lsh "
    # opcodes on atoms as vectors of bools 0x18-0x1c
    "logand logior logxor lognot . "
    # opcodes for bls 1381 0x1d-0x1f
    "point_add pubkey_for_exp . "
    # bool opcodes 0x20-0x23
    "not any all . "
    # misc 0x24
    "softfork "
).split()


KEYWORD_FROM_ATOM = {Program.int_to_bytes(k): v for k, v in enumerate(KWS)}
KEYWORD_TO_ATOM = {v: k for k, v in KEYWORD_FROM_ATOM.items()}


def format_pair(f, sexp: Program, keyword_from_atom):
    f.write("(")
    is_first = True
    while sexp.atom is None:
        pair = sexp.pair
        if not is_first:
            f.write(" ")
        format_program(f, pair[0], keyword_from_atom, is_first=is_first)
        sexp = pair[1]
        is_first = False
    if len(sexp.atom) > 0:
        f.write(" . ")
        format_program(f, sexp, keyword_from_atom, is_first=False)
    f.write(")")


def format_program(f, sexp: Program, keyword_from_atom, is_first=False):
    if sexp.pair:
        format_pair(f, sexp, keyword_from_atom)
        return

    atom = sexp.atom

    if is_first:
        kw = keyword_from_atom.get(atom)
        if kw is not None and kw != ".":
            f.write(kw)
            return

    type = type_for_atom(atom)
    assert type in "IHSD"
    p_table = dict(
        I=lambda a: "%d" % Program.int_from_bytes(a),
        H=lambda a: "0x%s" % a.hex(),
        D=lambda a: '"%s"' % a.decode("utf8"),
        S=lambda a: "'%s'" % a.decode("utf8"),
    )
    f.write(p_table[type](atom))


def disassemble(sexp, keyword_from_atom=KEYWORD_FROM_ATOM):
    f = io.StringIO()
    format_program(f, sexp, keyword_from_atom=keyword_from_atom)
    return f.getvalue()



from typing import Any, Callable, Dict, List, Tuple, TypeVar

from klvm_rs import Program


K = TypeVar("K")
T = TypeVar("T")
V = TypeVar("V")


def transform_dict(program, dict_transformer_f):
    """
    Transform elements of the dict d using the xformer (also a dict,
    where the keys match the keys in d and the values of d are transformed
    by invoking the corresponding values in xformer.
    """
    r = klvm_to_list(program, lambda x: dict_transformer_f(x.pair[0], x.pair[1]))
    d = dict(r)
    return d


def transform_by_key(
    key: Program,
    value: Program,
    transformation_lookup: Dict[str, Callable[[Program], Any]],
) -> Tuple[str, Any]:
    """
    Use this if the key is utf-8 and the value decoding depends on the key.
    """
    key_str = key.atom.decode()
    f = transformation_lookup.get(key_str, lambda x: x)
    final_value = f(value)
    return (key_str, final_value)


def transform_dict_by_key(
    transformation_lookup: Dict[str, Callable[[Program], Any]]
) -> Any:
    return lambda k, v: transform_by_key(k, v, transformation_lookup)


def transform_as_struct(items: Program, *struct_transformers) -> Tuple[Any, ...]:
    r = []
    for f in struct_transformers:
        r.append(f(items.pair[0]))
        items = items.pair[1]
    return tuple(r)


def klvm_to_list(
    item_list: Program, item_transformation_f: Callable[[Program], T]
) -> List[T]:
    r = []
    while item_list.pair:
        this_item, item_list = item_list.pair
        r.append(item_transformation_f(this_item))
    return r


def klvm_to_list_of_ints(items: Program) -> List[int]:
    return klvm_to_list(items, lambda obj: Program.to(obj).as_int())


def no_op(x):
    return x


def as_atom(x):
    return x.atom


def as_int(x):
    return x.as_int()



from dataclasses import dataclass

from typing import Dict, List

from chik_base.bls12_381 import BLSPublicKey, BLSSecretExponent

from hsmk.klvm_serde import Frugal


@dataclass
class SumHint(Frugal):
    public_keys: List[BLSPublicKey]
    synthetic_offset: BLSSecretExponent

    def final_public_key(self) -> BLSPublicKey:
        return sum(self.public_keys, start=self.synthetic_offset.public_key())


@dataclass
class PathHint(Frugal):
    root_public_key: BLSPublicKey
    path: List[int]

    def public_key(self) -> BLSPublicKey:
        return self.root_public_key.child_for_path(self.path)


PathHints = Dict[BLSPublicKey, PathHint]
SumHints = Dict[BLSPublicKey, SumHint]



from typing import List

import argparse

from chik_base.bls12_381 import BLSSignature
from chik_base.core import SpendBundle
from chik_base.cbincode import to_bytes

from hsmk.core.unsigned_spend import UnsignedSpend
from hsmk.process.sign import generate_synthetic_offset_signatures
from hsmk.util.qrint_encoding import a2b_qrint


def create_spend_bundle(unsigned_spend: UnsignedSpend, signatures: List[BLSSignature]):
    extra_signatures = generate_synthetic_offset_signatures(unsigned_spend)

    # now let's try adding them all together and creating a `SpendBundle`

    all_signatures = signatures + [sig_info.signature for sig_info in extra_signatures]
    total_signature = sum(all_signatures, start=all_signatures[0].zero())

    return SpendBundle(unsigned_spend.coin_spends, total_signature)


def file_or_string(p) -> str:
    try:
        with open(p) as f:
            text = f.read().strip()
    except Exception:
        text = p
    return text


def hsmkmerge(args, parser):
    blob = a2b_qrint(file_or_string(args.unsigned_spend))
    unsigned_spend = UnsignedSpend.from_bytes(blob)
    signatures = [
        BLSSignature.from_bytes(a2b_qrint(file_or_string(_))) for _ in args.signature
    ]
    spend_bundle = create_spend_bundle(unsigned_spend, signatures)
    print(to_bytes(spend_bundle).hex())


def create_parser():
    parser = argparse.ArgumentParser(
        description=(
            "Create a signed `SpendBundle` from `UnsignedSpends` " "and signatures."
        )
    )
    parser.add_argument(
        "unsigned_spend",
        metavar="path-to-unsigned-spend-as-hex",
        help="file containing hex-encoded `UnsignedSpends`",
    )
    parser.add_argument(
        "signature",
        metavar="qrint-encoded-signature",
        nargs="+",
        help="qrint-encoded signature",
    )
    return parser


def main():
    parser = create_parser()
    args = parser.parse_args()
    return hsmkmerge(args, parser)


if __name__ == "__main__":
    main()



import math
import zlib

from typing import List, Tuple


def optimal_chunk_size_for_max_chunk_size(full_size: int, max_chunk_size: int) -> int:
    payload_size = max_chunk_size - 2
    chunk_count = (full_size + payload_size - 1) // payload_size
    optimal_payload_size = (full_size + chunk_count - 1) // chunk_count
    optimal_chunk_size = optimal_payload_size + 2
    return optimal_chunk_size


def create_chunks_for_blob(blob: bytes, bytes_per_chunk: int) -> List[bytes]:
    total_len = len(blob)

    bytes_per_chunk -= 2
    num_chunks = math.ceil(total_len / bytes_per_chunk)
    if num_chunks > 256:
        raise ValueError("Cannot chunk a blob into more than 256 chunks")

    bundle_chunks = [
        blob[i : min(i + bytes_per_chunk, total_len)]
        + bytes(
            [int(i / bytes_per_chunk), num_chunks - 1]
        )  # Indication of the order of the blocks
        for i in range(0, total_len, bytes_per_chunk)
    ]

    return bundle_chunks


def chunks_for_zlib_blob(blob: bytes, bytes_per_chunk: int) -> List[bytes]:
    return create_chunks_for_blob(zlib.compress(blob, level=9), bytes_per_chunk)


class ChunkAssembler:
    chunks: List[bytes]

    def __init__(self, chunks=[]):
        self.chunks = []
        for chunk in chunks:
            self.add_chunk(chunk)

    def add_chunk(self, chunk: bytes):
        if chunk in self.chunks:
            return
        if len(self.chunks) > 0 and self.chunks[0][-1] != chunk[-1]:
            raise ValueError("chunk is part of a different set")
        indexes_seen = [c[-2] for c in self.chunks]
        if chunk[-2] in indexes_seen:
            raise ValueError("chunk conflicts with already added chunk")
        else:
            self.chunks.append(chunk)

    def is_assembled(self) -> bool:
        if len(self.chunks) > 0 and self.chunks[0][-1] == len(self.chunks) - 1:
            return True
        else:
            return False

    def status(self) -> Tuple[int, int]:
        """Returns: (amount of chunks we have, amount of chunks total)"""
        if len(self.chunks) == 0:
            return 0, 0
        else:
            return len(self.chunks), self.chunks[0][-1] + 1

    def __bytes__(self) -> bytes:
        if not self.is_assembled():
            raise ValueError("insufficient chunks")
        sorted_chunks = sorted(self.chunks, key=lambda c: c[-2])
        bare_chunks = [c[:-2] for c in sorted_chunks]
        return b"".join(bare_chunks)

    def assemble(self) -> bytes:
        return bytes(self)


def blob_for_chunks(chunks: List[bytes]) -> bytes:
    return ChunkAssembler(chunks).assemble()


def blob_for_zlib_chunks(chunks: List[bytes]) -> bytes:
    return zlib.decompress(blob_for_chunks(chunks))


assemble_chunks = blob_for_chunks



import sys

from chik_base.bls12_381 import BLSSecretExponent


def main(argv=sys.argv[1:]):
    for arg in argv:
        secret_exponent = BLSSecretExponent.from_bech32m(arg)
        print(secret_exponent.public_key().as_bech32m())


if __name__ == "__main__": # pragma: no cover
    main()
